from django.contrib.contenttypes.models import ContentType
from django.db.models import (
    BooleanField,
    Case,
    CharField,
    ExpressionWrapper,
    F,
    Q,
    Value,
    When,
)
from django.db.models.functions import Concat
from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from rest_framework import filters, permissions, status
from rest_framework.decorators import action
from rest_framework.response import Response
from reversion.views import RevisionMixin
from wbcore import viewsets
from wbcore.contrib.authentication.authentication import JWTCookieAuthentication
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.filters import DjangoFilterBackend

from wbcompliance.filters import ComplianceFormFilter, ComplianceFormSignatureFilter
from wbcompliance.models import (
    ComplianceForm,
    ComplianceFormRule,
    ComplianceFormSection,
    ComplianceFormSignature,
    ComplianceFormSignatureRule,
    ComplianceFormSignatureSection,
    ComplianceFormType,
    update_or_create_compliance_document,
)
from wbcompliance.serializers import (
    ComplianceFormModelSerializer,
    ComplianceFormRepresentationSerializer,
    ComplianceFormRuleModelSerializer,
    ComplianceFormSectionModelSerializer,
    ComplianceFormSectionRepresentationSerializer,
    ComplianceFormSignatureModelSerializer,
    ComplianceFormSignatureRuleModelSerializer,
    ComplianceFormSignatureSectionRepresentationSerializer,
    ComplianceFormTypeModelSerializer,
    ComplianceFormTypeRepresentationSerializer,
)

from .buttons import ComplianceFormButtonConfig, ComplianceFormSignatureButtonConfig
from .compliance_type import ComplianceType
from .display import (
    ComplianceFormDisplayConfig,
    ComplianceFormRuleDisplayConfig,
    ComplianceFormSectionDisplayConfig,
    ComplianceFormSignatureDisplayConfig,
    ComplianceFormSignatureSectionRuleDisplayConfig,
    ComplianceFormTypeDisplayConfig,
)
from .endpoints import (
    CFComplianceFormSectionEndpointConfig,
    CFComplianceFormSignatureEndpointConfig,
    ComplianceFormEndpointConfig,
    ComplianceFormRuleEndpointConfig,
    ComplianceFormSectionRuleEndpointConfig,
    ComplianceFormSignatureEndpointConfig,
    ComplianceFormSignatureSectionRuleEndpointConfig,
    ComplianceFormTypeEndpointConfig,
)
from .titles import (
    ComplianceFormSectionRuleTitleConfig,
    ComplianceFormSectionTitleConfig,
    ComplianceFormSignatureSectionRuleTitleConfig,
    ComplianceFormSignatureTitleConfig,
    ComplianceFormTitleConfig,
)


class ComplianceFormRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcompliance:complianceformrepresentation"

    queryset = ComplianceForm.objects.all()
    serializer_class = ComplianceFormRepresentationSerializer


class ComplianceFormSectionRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcompliance:complianceformsectionrepresentation"

    queryset = ComplianceFormSection.objects.all()
    serializer_class = ComplianceFormSectionRepresentationSerializer


class ComplianceFormSignatureSectionRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcompliance:complianceformsignaturesectionrepresentation"

    queryset = ComplianceFormSignatureSection.objects.all()
    serializer_class = ComplianceFormSignatureSectionRepresentationSerializer


# TYPE OF THE COMPLIANCE FORM
class ComplianceFormTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcompliance:complianceformtyperepresentation"
    queryset = ComplianceFormType.objects.all()
    serializer_class = ComplianceFormTypeRepresentationSerializer


class ComplianceFormTypeViewSet(RevisionMixin, viewsets.ModelViewSet):
    endpoint_config_class = ComplianceFormTypeEndpointConfig
    display_config_class = ComplianceFormTypeDisplayConfig

    ordering_fields = ["name"]

    serializer_class = ComplianceFormTypeModelSerializer
    queryset = ComplianceFormType.objects.all()

    def get_queryset(self):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_queryset()
        return ComplianceFormType.objects.none()


# SECTION OF THE COMPLIANCE FORM
class ComplianceFormSectionViewSet(RevisionMixin, viewsets.ModelViewSet):
    title_config_class = ComplianceFormSectionTitleConfig
    display_config_class = ComplianceFormSectionDisplayConfig

    search_fields = ["name"]
    ordering_fields = ["name"]
    queryset = ComplianceFormSection.objects.all()
    serializer_class = ComplianceFormSectionModelSerializer

    def get_queryset(self):
        if ComplianceType.is_administrator(self.request.user):
            return ComplianceFormSection.objects.select_related("compliance_form")
        return ComplianceFormSection.objects.none()


class ComplianceFormSectionComplianceFormViewSet(ComplianceFormSectionViewSet):
    endpoint_config_class = CFComplianceFormSectionEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(compliance_form__id=self.kwargs["compliance_form_id"])


# RULES
class ComplianceFormRuleViewSet(RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbcompliance:complianceformrule"
    display_config_class = ComplianceFormRuleDisplayConfig
    endpoint_config_class = ComplianceFormRuleEndpointConfig

    filter_backends = (
        DjangoFilterBackend,
        filters.OrderingFilter,
    )

    serializer_class = ComplianceFormRuleModelSerializer

    queryset = ComplianceFormRule.objects.select_related("section")


# RULES OF THE SECTION OF THE COMPLIANCE FORM
class ComplianceFormSectionRuleViewSet(ComplianceFormRuleViewSet):
    IDENTIFIER = "wbcompliance:complianceformsection-rules"
    title_config_class = ComplianceFormSectionRuleTitleConfig
    endpoint_config_class = ComplianceFormSectionRuleEndpointConfig

    def get_queryset(self):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_queryset().filter(section__id=self.kwargs["section_id"])
        return ComplianceFormRule.objects.none()


# RULES OF THE SECTION OF THE COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureSectionRuleViewSet(RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbcompliance:complianceformsignaturesection-signaturerules"
    title_config_class = ComplianceFormSignatureSectionRuleTitleConfig
    display_config_class = ComplianceFormSignatureSectionRuleDisplayConfig
    endpoint_config_class = ComplianceFormSignatureSectionRuleEndpointConfig

    serializer_class = ComplianceFormSignatureRuleModelSerializer

    queryset = ComplianceFormSignatureRule.objects.all()

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(section__id=self.kwargs["section_id"])
            .annotate(
                expected_result=ComplianceFormSignatureRule.get_subquery_expected_ticked(),
                same_answer=ExpressionWrapper(
                    Q(ticked__exact=F("expected_result")),
                    output_field=BooleanField(),
                ),
            )
        )


# COMPLIANCE FORM
class ComplianceFormModelViewSet(RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbcompliance:complianceform"

    ordering_fields = [
        "title",
        "status",
        "creator",
        "changer",
        "changed",
        "start",
        "end",
        "version",
        "form_type",
        "compliance_type",
    ]
    ordering = ["-changed", "-created"]
    search_fields = ["title"]

    filterset_class = ComplianceFormFilter

    display_config_class = ComplianceFormDisplayConfig
    title_config_class = ComplianceFormTitleConfig
    button_config_class = ComplianceFormButtonConfig
    endpoint_config_class = ComplianceFormEndpointConfig

    serializer_class = ComplianceFormModelSerializer
    queryset = ComplianceForm.objects.select_related(
        "creator",
        "changer",
        "form_type",
        "compliance_type",
    ).prefetch_related("assigned_to")

    @cached_property
    def user_has_compliance_admin_permission(self):
        return self.request.user.has_perm("wbcompliance.administrate_compliance")

    def get_queryset(self):
        if ComplianceType.is_administrator(self.request.user):
            qs = super().get_queryset()
            qs = qs.annotate(
                # is_signed = ComplianceForm.get_subquery_compliance_form_signature(self.request.user.profile),
                total_compliance_form_signature=ComplianceForm.get_subquery_total_compliance_form_signature(),
                total_signed=F("total_compliance_form_signature")
                - ComplianceForm.get_subquery_total_compliance_form_signature(remaining_signed=True),
                current_signed=Concat(
                    Value(" ("),
                    F("total_signed"),
                    Value("/"),
                    F("total_compliance_form_signature"),
                    Value(")"),
                    output_field=CharField(),
                ),
                is_signed=Case(
                    When(total_compliance_form_signature=F("total_signed"), then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                ),
            )
        else:
            qs = super().get_queryset().filter(status=ComplianceForm.Status.ACTIVE)
            qs = qs.annotate(
                is_signed=ComplianceForm.get_subquery_compliance_form_signature(self.request.user.profile),
            ).filter(is_signed__isnull=False)
        return qs

    @action(detail=True, methods=["PATCH"], permission_classes=[permissions.IsAuthenticated])
    def sendcomplianceformnotification(self, request, pk=None):
        compliance_form = ComplianceForm.objects.get(pk=pk)
        person_ids = ComplianceFormSignature.objects.filter(
            compliance_form=compliance_form, version=compliance_form.version, signed=None
        ).values_list("person", flat=True)

        if person_ids:
            msg = _("<p>{} has activated a {} {} version {} </p>").format(
                str(compliance_form.changer), compliance_form.title, compliance_form.form_type, compliance_form.version
            )
            if compliance_form.policy and compliance_form.policy != "<p></p>" and compliance_form.policy != "null":
                msg += _("</br><p><b>User's ComplianceForm:</b></p><i>{}</i>").format(compliance_form.policy)
            title = _("Unsigned Compliance Form - {} Reminder : {} ").format(
                compliance_form.form_type, compliance_form.title
            )
            compliance_form.notify(title, msg, recipients=User.objects.filter(profile__id__in=person_ids))

            return Response(
                {
                    "__notification": {
                        "compliance_form": compliance_form.id,
                        "status": "Compliance Form notification sent",
                        "person_signed": person_ids,
                    }
                },
                status=status.HTTP_200_OK,
            )
        return Response(
            {
                "__notification": {
                    "compliance_form": compliance_form.id,
                    "status": "Compliance Form notification not sent",
                    "person_signed": None,
                }
            },
            status=status.HTTP_400_BAD_REQUEST,
        )

    @action(detail=True, methods=["PATCH"], authentication_classes=[JWTCookieAuthentication])
    def regenerate_document(self, request, pk):
        if not request.user.has_perm("wbcompliance.administrate_compliance"):
            return Response(status=status.HTTP_403_FORBIDDEN)
        content_type = ContentType.objects.get_for_model(ComplianceForm)
        send_email = request.POST.get("send_email", "false") == "true"
        update_or_create_compliance_document.delay(request.user.id, content_type.id, pk, send_email=send_email)

        return Response(
            {"__notification": {"title": "PDF is going to be created and sent to you."}},
            status=status.HTTP_200_OK,
        )


# COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureModelViewSet(RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbcompliance:complianceformsignature"
    display_config_class = ComplianceFormSignatureDisplayConfig
    title_config_class = ComplianceFormSignatureTitleConfig
    button_config_class = ComplianceFormSignatureButtonConfig
    endpoint_config_class = ComplianceFormSignatureEndpointConfig

    search_fields = ["compliance_form__title"]

    ordering_fields = ["version", "signed", "start", "end", "person"]
    ordering = ["-version", "-signed"]

    filterset_class = ComplianceFormSignatureFilter

    serializer_class = ComplianceFormSignatureModelSerializer
    queryset = ComplianceFormSignature.objects.select_related("compliance_form", "person")

    def get_queryset(self):
        if ComplianceType.is_administrator(self.request.user):
            qs = super().get_queryset()
        else:
            qs = super().get_queryset().filter(person=self.request.user.profile)
        return qs.annotate(
            is_signed=Case(
                When(signed=None, then=Value(False)),
                default=Value(True),
                output_field=BooleanField(),
            )
        )

    @action(detail=True, methods=["PATCH"], permission_classes=[permissions.IsAuthenticated])
    def signature(self, request, pk=None):
        compliance_form_signature = get_object_or_404(ComplianceFormSignature, pk=pk)
        if request.user.profile != compliance_form_signature.person:
            return Response(status=status.HTTP_403_FORBIDDEN)
        compliance_form_signature.signed = timezone.now()
        compliance_form_signature.remark = request.POST.get("remark")
        compliance_form_signature.save()

        ComplianceFormSignature.objects.filter(
            Q(compliance_form=compliance_form_signature.compliance_form)
            & Q(version__lt=compliance_form_signature.version)
            & Q(signed__isnull=True)
            & Q(person=request.user.profile)
        ).update(signed=timezone.now())

        if hasattr(compliance_form_signature.person, "user_account"):
            msg = _("<p><b>{}</b> {} version <b>{} </b> has been signed</p>").format(
                compliance_form_signature.compliance_form.title,
                compliance_form_signature.compliance_form.form_type,
                compliance_form_signature.version,
            )
            if (
                compliance_form_signature.policy
                and compliance_form_signature.policy != "<p></p>"
                and compliance_form_signature.policy != "null"
            ):
                msg += _("</br><p><b>{}:</b></p> <i>{}</i>").format(
                    compliance_form_signature.compliance_form.form_type, compliance_form_signature.policy
                )
            send_notification(
                code="wbcompliance.complianceformsignature.signed",
                title=_("Confirmation : {} {} version {} has been signed").format(
                    compliance_form_signature.compliance_form.title,
                    compliance_form_signature.compliance_form.form_type,
                    compliance_form_signature.version,
                ),
                body=msg,
                user=compliance_form_signature.person.user_account,
                reverse_name="wbcompliance:complianceformsignature-detail",
                reverse_args=[compliance_form_signature.id],
            )

        return Response(
            {
                "__notification": {
                    str(compliance_form_signature.signed): "Compliance Form signed",
                    "person_signed": compliance_form_signature.person.id,
                }
            },
            status=status.HTTP_200_OK,
        )

    @action(detail=True, methods=["PATCH"], authentication_classes=[JWTCookieAuthentication])
    def regenerate_document(self, request, pk):
        compliance_form_signature = get_object_or_404(ComplianceFormSignature, pk=pk)
        if (
            not request.user.has_perm("wbcompliance.administrate_compliance")
            and request.user.profile != compliance_form_signature.person
        ):
            return Response(status=status.HTTP_403_FORBIDDEN)
        content_type = ContentType.objects.get_for_model(ComplianceFormSignature)
        send_email = request.POST.get("send_email", "false") == "true"
        update_or_create_compliance_document.delay(request.user.id, content_type.id, pk, send_email=send_email)

        return Response(
            {"__notification": {"title": "PDF is going to be created and sent to you."}},
            status=status.HTTP_200_OK,
        )


class CFComplianceFormSignatureModelViewSet(ComplianceFormSignatureModelViewSet):
    endpoint_config_class = CFComplianceFormSignatureEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(compliance_form__id=self.kwargs["compliance_form_id"])



from django.dispatch import receiver
from wbcore.test.signals import (
    custom_update_data_from_factory,
    custom_update_kwargs,
    get_custom_factory,
)

from wbcompliance.factories import (
    ComplianceTaskInstanceReviewFactory,
    ComplianceTaskInstanceReviewNoGroupFactory,
    ComplianceTaskReviewFactory,
    ComplianceTaskReviewNoGroupFactory,
)
from wbcompliance.viewsets import (
    CFComplianceFormSignatureModelViewSet,
    ComplianceFormSignatureModelViewSet,
    ComplianceFormSignatureSectionRuleViewSet,
    ComplianceTaskInstanceReviewGroupModelViewSet,
    ComplianceTaskInstanceReviewNoGroupModelViewSet,
    ComplianceTaskReviewGroupModelViewSet,
    ComplianceTaskReviewNoGroupModelViewSet,
)

# =================================================================================================================
#                                              CUSTOM FACTORY
# =================================================================================================================


@receiver(get_custom_factory, sender=ComplianceTaskReviewNoGroupModelViewSet)
def receive_factory_compliance_task_review_no_group(sender, *args, **kwargs):
    return ComplianceTaskReviewNoGroupFactory


@receiver(get_custom_factory, sender=ComplianceTaskReviewGroupModelViewSet)
def receive_factory_compliance_task_review(sender, *args, **kwargs):
    return ComplianceTaskReviewFactory


@receiver(get_custom_factory, sender=ComplianceTaskInstanceReviewNoGroupModelViewSet)
def receive_factory_compliance_task_instance_review_no_group(sender, *args, **kwargs):
    return ComplianceTaskInstanceReviewNoGroupFactory


@receiver(get_custom_factory, sender=ComplianceTaskInstanceReviewGroupModelViewSet)
def receive_factory_compliance_task_review_group(sender, *args, **kwargs):
    return ComplianceTaskInstanceReviewFactory


# =================================================================================================================
#                                              UPDATE DATA
# =================================================================================================================


@receiver(custom_update_data_from_factory, sender=ComplianceFormSignatureSectionRuleViewSet)
def receive_data_compliance_form_signature_section(sender, *args, **kwargs):
    if (obj := kwargs.get("obj_factory")) and (user := kwargs.get("user")):
        obj.section.compliance_form_signature.person = user.profile
        obj.section.compliance_form_signature.save()
    return {}


@receiver(custom_update_data_from_factory, sender=ComplianceFormSignatureModelViewSet)
@receiver(custom_update_data_from_factory, sender=CFComplianceFormSignatureModelViewSet)
def receive_data_compliance_form_signature(sender, *args, **kwargs):
    if (obj := kwargs.get("obj_factory")) and (user := kwargs.get("user")):
        obj.person = user.profile
        obj.signed = None
        obj.save()
    return {}


# =================================================================================================================
#                                              UPDATE KWARGS
# =================================================================================================================


@receiver(custom_update_kwargs, sender=ComplianceTaskInstanceReviewNoGroupModelViewSet)
@receiver(custom_update_kwargs, sender=ComplianceTaskReviewGroupModelViewSet)
@receiver(custom_update_kwargs, sender=ComplianceTaskReviewNoGroupModelViewSet)
def receive_kwargs_compliance_task_review(sender, *args, **kwargs):
    if (obj := kwargs.get("obj_factory")) and obj.review.exists():
        return {"review_id": obj.review.first().id}
    return {}


@receiver(custom_update_kwargs, sender=ComplianceTaskInstanceReviewGroupModelViewSet)
def receive_kwargs_compliance_task_instance_review(sender, *args, **kwargs):
    if (obj := kwargs.get("obj_factory")) and obj.review.exists():
        return {"review_id": obj.review.first().id, "group_id": obj.task.group.id}
    return {}



from .checks import RiskCheckDisplayConfig
from .incidents import (
    CheckedObjectIncidentRelationshipRiskRuleDisplayConfig,
    RiskIncidentDisplayConfig,
)
from .rules import (
    RiskRuleDisplayConfig,
    RuleCheckedObjectRelationshipRiskRuleDisplayConfig,
    RuleThresholdRiskRuleDisplayConfig,
)
from .tables import RiskManagementIncidentTableDisplayConfig



# Generated by Django 5.0.9 on 2024-10-17 07:58

from django.db import migrations
from django.db.models import Count


def remove_duplicates(apps, schema_editor):
    RuleCheckedObjectRelationship = apps.get_model("wbcompliance", "RuleCheckedObjectRelationship")
    for row in (
        RuleCheckedObjectRelationship.objects.values("rule", "checked_object_content_type", "checked_object_id")
        .annotate(c=Count("*"))
        .filter(c__gt=1)
    ):
        qs = RuleCheckedObjectRelationship.objects.filter(
            rule=row["rule"],
            checked_object_content_type=row["checked_object_content_type"],
            checked_object_id=row["checked_object_id"],
        )
        base = qs.first()
        for rel in qs.exclude(id=base.id):
            rel.checks.update(rule_checked_object_relationship=rel)
            rel.delete()


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("wbcompliance", "0017_alter_rulebackend_incident_report_template"),
    ]

    operations = [
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(remove_duplicates),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterUniqueTogether(
            name="rulecheckedobjectrelationship",
            unique_together={("rule", "checked_object_content_type", "checked_object_id")},
        ),
    ]



from contextlib import suppress
from datetime import date, timedelta
from types import DynamicClassAttribute
from typing import Any, Self

from celery import shared_task
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.template import Context, Template
from django.utils.translation import gettext_lazy as _
from pandas._libs.tslibs.offsets import BDay
from psycopg.types.range import DateRange
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin
from wbcore.workers import Queue

from .incidents import CheckedObjectIncidentRelationship, RiskIncident, RiskIncidentType


class RiskCheckManager(models.Manager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                passive_check=models.Case(
                    models.When(
                        models.Q(activator_id__isnull=True) & models.Q(activator_content_type__isnull=True),
                        then=models.Value(True),
                    ),
                    default=models.Value(False),
                )
            )
        )


class PassiveRiskCheckManager(RiskCheckManager):
    def get_queryset(self):
        return super().get_queryset().filter(passive_check=True)


class RiskCheck(ComplexToStringMixin, WBModel):
    class CheckStatus(models.TextChoices):
        PENDING = "PENDING", "Pending"
        RUNNING = "RUNNING", "Running"
        FAILED = "FAILED", "Failed"
        SUCCESS = "SUCCESS", "Success"
        WARNING = "WARNING", "Warning"

        @DynamicClassAttribute
        def icon(self):
            return {
                "PENDING": WBIcon.PENDING.icon,
                "RUNNING": WBIcon.RUNNING.icon,
                "FAILED": WBIcon.REJECT.icon,
                "SUCCESS": WBIcon.CONFIRM.icon,
                "WARNING": WBIcon.WARNING.icon,
            }[self.value]

        @DynamicClassAttribute
        def color(self):
            return {
                "PENDING": WBColor.YELLOW_LIGHT.value,
                "RUNNING": WBColor.BLUE_LIGHT.value,
                "FAILED": WBColor.RED_LIGHT.value,
                "SUCCESS": WBColor.GREEN_LIGHT.value,
                "WARNING": WBColor.YELLOW_DARK.value,
            }[self.value]

    rule = models.ForeignKey(to="wbcompliance.RiskRule", related_name="checks", on_delete=models.CASCADE)

    creation_datetime = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Creation Date"),
        help_text=_("Time at which the check was created/triggered"),
    )
    evaluation_date = models.DateField(
        verbose_name=_("Evaluation Date"), help_text=_("The date at which the rule was evaluated")
    )

    checked_object_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="triggered_checks"
    )
    checked_object_id = models.PositiveIntegerField()
    checked_object = GenericForeignKey("checked_object_content_type", "checked_object_id")
    checked_object_repr = models.CharField(max_length=256, blank=True, null=True)

    status = models.CharField(
        max_length=32, default=CheckStatus.PENDING, choices=CheckStatus.choices, verbose_name=_("Status")
    )
    activator_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, max_length=256, blank=True, null=True
    )
    activator_id = models.PositiveIntegerField(blank=True, null=True)
    activator = GenericForeignKey("activator_content_type", "activator_id")

    objects = PassiveRiskCheckManager()
    all_objects = RiskCheckManager()

    @property
    def needs_incident(self) -> bool:
        # if the there is a similar checked object relationship, we needs to create a general incident if the check evaluates to a breach
        return (
            self.rule.checked_object_relationships.filter(
                checked_object_content_type=self.checked_object_content_type, checked_object_id=self.checked_object_id
            ).exists()
            or self.precheck
        )

    @property
    def precheck(self) -> bool:
        return self.activator_id is not None

    @property
    def previous_check(self) -> Self | None:
        """
        Property holding the last valid check

        Returns:
            The last valid check
        """
        with suppress(RiskCheck.DoesNotExist):
            return (
                RiskCheck.objects.filter(
                    evaluation_date__lt=self.evaluation_date,
                    rule=self.rule,
                    checked_object_content_type=self.checked_object_content_type,
                    checked_object_id=self.checked_object_id,
                )
                .order_by("-evaluation_date", "-creation_datetime")
                .first()
            )

    def save(self, *args, **kwargs):
        self.checked_object_repr = str(self.checked_object)
        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        incidents = RiskIncident.all_objects.filter(
            id__in=list(
                CheckedObjectIncidentRelationship.objects.filter(rule_check_id=self.id).values_list(
                    "incident", flat=True
                )
            )
        )
        super().delete(*args, **kwargs)
        # ensure that incident without subincidents are deleted automatically
        for incident in incidents:
            if not incident.checked_object_relationships.exists():
                incident.delete()

    def compute_str(self) -> str:
        return _("{} - {}").format(
            self.checked_object,
            self.evaluation_date,
        )

    def evaluate(
        self, *explicit_dto, override_incident: bool = False, ignore_informational_threshold: bool = False, **kwargs
    ) -> list[models.Model]:
        """
        Evaluate the check and returns tuple of incidents information
        Args:
            override_incident: True if the existing incident needs to be overriden

        Returns:

        """
        self.status = self.CheckStatus.RUNNING
        self.save()
        rule_backend = self.rule.rule_backend.backend(
            self.evaluation_date, self.checked_object, self.rule.parameters, self.rule.thresholds.all()
        )
        self.status = self.CheckStatus.SUCCESS
        incidents = []
        report_template = Template(self.rule.rule_backend.incident_report_template)
        for incident_result in rule_backend.check_rule(*explicit_dto, **kwargs):
            if (
                ignore_informational_threshold
                and incident_result.severity.is_ignorable
                and incident_result.severity.is_informational
            ):
                self.status = self.CheckStatus.WARNING
            else:
                self.status = self.CheckStatus.FAILED
                # If the check is passive, we aggregated incident per breached object and return it for further processing
            if self.needs_incident:
                report = report_template.render(Context({"report_details": incident_result.report_details}))
                incident, created = self.get_or_create_incident(
                    self.evaluation_date,
                    incident_result.severity,
                    incident_result.breached_object,
                    incident_result.breached_object_repr,
                )
                incident.update_or_create_relationship(
                    self,
                    report,
                    incident_result.report_details,
                    incident_result.breached_value,
                    incident_result.severity,
                    override_incident=override_incident or created,
                )
                incidents.append(incident)
            else:
                # If the check is active, the only thing that matter is whether the check led to incident or not
                report = f"<p><strong>{incident_result.breached_object_repr}</strong>: {incident_result.breached_value}</p>"  # as we do not store the breached object, we need a way to display it in the subincident relationship. This is not optimal though
                CheckedObjectIncidentRelationship.objects.create(
                    rule_check=self,
                    report=report,
                    report_details=incident_result.report_details,
                    breached_value=incident_result.breached_value,
                    severity=incident_result.severity,
                )
        self.save()
        return incidents

    def get_or_create_incident(
        self,
        evaluation_date: date,
        incident_severity: "RiskIncidentType",
        breached_object: Any,
        breached_object_repr: str,
    ) -> tuple[RiskIncident, bool]:
        """
        Utility function to get or create incident base on the given breached object

        Args:
            evaluation_date: The incident date
            incident_severity: The incident severity
            breached_object: The breached object (i.e. the object that triggers the incident)
            breached_object_repr: Its string representation

        Returns:
            A tuple (incident, is_created)
        """

        # Consider that if an incident happens one business day in the future or past, it is continue
        date_range = DateRange(  # type: ignore
            (evaluation_date - BDay(1)).date(), (evaluation_date + BDay(1)).date(), "[]"
        )
        incidents = RiskIncident.objects.filter(rule=self.rule, date_range__overlap=date_range, precheck=self.precheck)
        if breached_object:
            # If a breached_object is provided, we lookup over it
            incidents = incidents.filter(
                breached_content_type=ContentType.objects.get_for_model(breached_object),
                breached_object_id=breached_object.id,
            )
        else:
            # Otherwise, we lookup over its string representation. Might need to change as this is not very robust
            incidents = incidents.filter(
                breached_object_repr=breached_object_repr,
            )
        created = False
        if incident := incidents.first():
            incident.date_range = DateRange(  # type: ignore
                lower=incident.date_range.lower, upper=(evaluation_date + timedelta(days=1))
            )
            incident.severity = incident_severity
        else:
            incident = RiskIncident(
                rule=self.rule,
                date_range=DateRange(lower=evaluation_date, upper=(evaluation_date + timedelta(days=1))),  # type: ignore
                breached_content_type=ContentType.objects.get_for_model(breached_object) if breached_object else None,
                breached_object_id=breached_object.id if breached_object else None,
                breached_object_repr=breached_object_repr,
                severity=incident_severity,
                status=RiskIncident.Status.OPEN,
            )
            created = True

        if self.rule.automatically_close_incident:
            incident.status = RiskIncident.Status.CLOSED
        elif self.precheck:
            incident.status = RiskIncident.Status.IGNORED
        incident.precheck = self.precheck
        incident.save()
        return incident, created

    class Meta:
        verbose_name = "Risk Check"
        verbose_name_plural = "Risk Checks"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:riskcheck"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:riskcheckrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


@shared_task(queue=Queue.HIGH_PRIORITY.value)
def evaluate_as_task(
    check_id: int,
    *dto,
    override_incident: bool = False,
    ignore_informational_threshold: bool = False,
    ignore_breached_object_content_types: list[tuple[int, int]] | None = None,
):
    if ignore_breached_object_content_types:
        ignore_breached_objects = [
            ContentType.objects.get(id=r[1]).get_object_for_this_type(id=r[0])
            for r in ignore_breached_object_content_types
        ]
    else:
        ignore_breached_objects = []
    check = RiskCheck.all_objects.get(id=check_id)
    check.evaluate(
        *dto,
        override_incident=override_incident,
        ignore_informational_threshold=ignore_informational_threshold,
        ignore_breached_objects=ignore_breached_objects,
    )



import json
import random
from datetime import date

import factory
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from faker import Faker
from guardian.utils import get_anonymous_user
from psycopg.types.range import NumericRange
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models.users import User

from wbcompliance.models.risk_management.rules import (
    RiskRule,
    RuleBackend,
    RuleCheckedObjectRelationship,
    RuleThreshold,
)

fake = Faker()


class RuleBackendFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("name")
    backend_class_path = "wbcompliance.factories.risk_management.backends"
    backend_class_name = "RuleBackend"
    allowed_checked_object_content_type = None

    class Meta:
        model = RuleBackend


def _get_default_parameter():
    return json.loads(
        json.dumps(
            {
                "date": fake.date_object(),
                "name": fake.name(),
                "int": fake.pyint(),
                "anonymous_user": get_anonymous_user().id,
            },
            cls=DjangoJSONEncoder,
        )
    )


class RiskRuleFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("name")
    description = factory.Faker("paragraph")
    rule_backend = factory.SubFactory("wbcompliance.factories.risk_management.RuleBackendFactory")
    is_enable = True
    only_passive_check_allowed = True
    is_silent = False
    is_mandatory = True
    automatically_close_incident = False
    parameters = factory.LazyAttribute(lambda x: _get_default_parameter())
    activation_date = date.min

    @factory.post_generation
    def threshold(self, create, extracted, **kwargs):
        if not RuleThreshold.objects.filter(rule=self).exists():
            RuleThresholdFactory.create(rule=self, notifiable_users=[UserFactory.create().profile])

    class Meta:
        model = RiskRule
        skip_postgeneration_save = True


class RuleThresholdFactory(factory.django.DjangoModelFactory):
    rule = factory.SubFactory("wbcompliance.factories.risk_management.RiskRuleFactory")
    range = factory.LazyAttribute(lambda o: NumericRange(lower=random.uniform(0, 0.5), upper=random.uniform(0.5, 1)))
    severity = factory.SubFactory("wbcompliance.factories.risk_management.incidents.RiskIncidentTypeFactory")

    @factory.post_generation
    def notifiable_users(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for person in extracted:
                self.notifiable_users.add(person)

    @factory.post_generation
    def notifiable_groups(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for group in extracted:
                self.notifiable_groups.add(group)

    class Meta:
        model = RuleThreshold
        skip_postgeneration_save = True


class RuleCheckedObjectRelationshipFactory(factory.django.DjangoModelFactory):
    rule = factory.SubFactory("wbcompliance.factories.risk_management.RiskRuleFactory")
    checked_object_content_type = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(User))
    checked_object_id = 1

    class Meta:
        model = RuleCheckedObjectRelationship
        django_get_or_create = ["rule", "checked_object_content_type", "checked_object_id"]



from contextlib import suppress

from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, Menu, MenuItem

from wbcompliance.models.risk_management.rules import RuleGroup

RULE_MENUITEM = MenuItem(
    label=_("Rules"),
    endpoint="wbcompliance:riskrule-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbcompliance.view_riskrule"]
    ),
    add=MenuItem(
        label=_("Add Rule"),
        endpoint="wbcompliance:riskrule-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbcompliance.add_riskrule"]
        ),
    ),
)


def _get_incident_menu():
    items = []
    with suppress(Exception):
        for rule_group in RuleGroup.objects.all():
            items.append(
                MenuItem(
                    label=rule_group.name,
                    endpoint="wbcompliance:riskincident-list",
                    permission=ItemPermission(
                        method=lambda request: request.user.is_internal,
                        permissions=["wbcompliance.view_riskincident"],
                    ),
                    endpoint_get_parameters={"rule_group": rule_group.id},
                )
            )
    if items:
        return Menu(label=_("Incidents"), items=items)
    return MenuItem(
        label=_("Incidents"),
        endpoint="wbcompliance:riskincident-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbcompliance.view_riskincident"]
        ),
    )


RISK_INCIDENT_MENUITEM = _get_incident_menu()

RISKINCIDENTTABLE_MENUITEM = MenuItem(
    label=_("Incident Table"),
    endpoint="wbcompliance:riskmanagementincidenttable-list",
    permission=ItemPermission(permissions=["wbcompliance.administrate_riskrule"]),
)



from datetime import timedelta

import pandas as pd
import pytest
from faker import Faker
from pandas.tseries.offsets import BDay

fake = Faker()


@pytest.mark.django_db
class TestChecks:
    def test_init(self, risk_check):
        assert risk_check

    @pytest.mark.parametrize("evaluation_date", [fake.date_object()])
    def test_previous_check(self, risk_rule, risk_check_factory, rule_checked_object_relationship, evaluation_date):
        check1 = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship, evaluation_date=evaluation_date
        )
        assert not check1.previous_check
        check2 = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship, evaluation_date=evaluation_date
        )
        check3 = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship,
            evaluation_date=evaluation_date + timedelta(days=1),
        )

        assert check2.previous_check is None
        assert check3.previous_check == check2
        # Shouldn't include this check in the queryset
        other_check = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship,
            evaluation_date=evaluation_date + timedelta(days=1),
        )
        assert other_check.previous_check == check2

    def test_get_unchecked_dates(
        self, weekday, risk_rule, risk_check_factory, rule_checked_object_relationship_factory
    ):
        rel = rule_checked_object_relationship_factory.create(rule=risk_rule)
        maximum_day_interval = 10
        assert list(rel.get_unchecked_dates(to_date=weekday, maximum_day_interval=maximum_day_interval)) == list(
            map(lambda _d: _d.date(), pd.date_range(weekday - timedelta(days=maximum_day_interval), weekday, freq="B"))
        )
        risk_check = risk_check_factory.create(rule_checked_object_relationship=rel, evaluation_date=weekday)

        assert list(rel.get_unchecked_dates()) == [(risk_check.evaluation_date + BDay(1)).date()]
        assert list(rel.get_unchecked_dates(to_date=risk_check.evaluation_date)) == list()

        risk_check2 = risk_check_factory.create(
            rule_checked_object_relationship=rel, evaluation_date=risk_check.evaluation_date + BDay(1)
        )
        assert list(rel.get_unchecked_dates()) == [(risk_check2.evaluation_date + BDay(1)).date()]



# Generated by Django 4.2.10 on 2024-03-06 09:29

import django.core.validators
import wbcore.models.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcompliance", "0013_alter_riskrule_unique_together"),
    ]

    operations = [
        migrations.AlterField(
            model_name="reviewcompliancetask",
            name="year",
            field=wbcore.models.fields.YearField(
                blank=True,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(1000),
                    django.core.validators.MaxValueValidator(9999),
                ],
                verbose_name="Year",
            ),
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-26 13:47

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcompliance", "0012_alter_compliancetype_options"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="riskrule",
            unique_together=set(),  # we cannot have a uniqueness on these two fields because when the rule is created, it set the parameters with the default value until it is changed in a second step by the user. Furthermore, we kind of want user to be able to create similar rule but with different threashold and notifiee policy
        ),
    ]



# Generated by Django 5.2.9 on 2025-12-16 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('directory', '0014_alter_entry_relationship_managers_and_more'),
        ('wbcompliance', '0022_riskincident_precheck'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='checkedobjectincidentrelationship',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='rulecheckedobjectrelationship',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='rulethreshold',
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name='checkedobjectincidentrelationship',
            constraint=models.UniqueConstraint(fields=('incident', 'rule_check'), name='unique_incident'),
        ),
        migrations.AddConstraint(
            model_name='rulecheckedobjectrelationship',
            constraint=models.UniqueConstraint(fields=('rule', 'checked_object_content_type', 'checked_object_id'), name='unique_checkedobjectrulerelationship'),
        ),
        migrations.AddConstraint(
            model_name='rulethreshold',
            constraint=models.UniqueConstraint(fields=('rule', 'severity'), name='unique_rule'),
        ),
    ]



from django.db.models import Max, Q
from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters

from wbcompliance.models import (
    ComplianceAction,
    ComplianceEvent,
    ComplianceForm,
    ComplianceFormSignature,
    ComplianceTask,
    ComplianceTaskGroup,
    ComplianceTaskInstance,
    ComplianceType,
    ReviewComplianceTask,
)


class ComplianceFormFilter(wb_filters.FilterSet):
    is_signed = wb_filters.BooleanFilter(label=_("Is Signed"), required=False)

    class Meta:
        model = ComplianceForm
        fields = {
            "form_type": ["exact"],
            "compliance_type": ["exact"],
            "title": ["exact", "icontains"],
            "creator": ["exact"],
            "changer": ["exact"],
            "version": ["exact"],
            "status": ["exact"],
            "start": ["gte", "lte"],
            "end": ["gte", "lte"],
        }


class ComplianceFormSignatureFilter(wb_filters.FilterSet):
    is_signed = wb_filters.BooleanFilter(label=_("Is Signed"), required=False)
    last_version = wb_filters.BooleanFilter(label=_("Last version"), initial=True, method="filter_version")
    # compliance_form = wb_filters.ModelChoiceFilter(
    #     label=_("Compliance Form"),
    #     queryset=ComplianceForm.objects.filter(status=ComplianceForm.Status.ACTIVE),
    #     endpoint=ComplianceForm.get_representation_endpoint(),
    #     value_key=ComplianceForm.get_representation_value_key(),
    #     label_key = ComplianceForm.get_representation_label_key(),
    #     method="filter_compliance_form"
    # )
    # def filter_compliance_form(self, queryset, name, value):
    #     return queryset.filter(status=ComplianceForm.Status.ACTIVE)

    class Meta:
        model = ComplianceFormSignature
        fields = {
            "person": ["exact"],
            "signed": ["gte", "lte"],
            "version": ["exact"],
            "compliance_form": ["exact"],
        }

    def filter_version(self, queryset, name, value):
        if value:
            max_compliance_forms = queryset.values("compliance_form").annotate(max_version=Max("version")).order_by()
            q_statement = Q()
            for max_compliance_form in max_compliance_forms:
                q_statement |= Q(compliance_form__exact=max_compliance_form["compliance_form"]) & Q(
                    version=max_compliance_form["max_version"]
                )
            return queryset.filter(q_statement)

        return queryset


class ComplianceTypeFilter(wb_filters.FilterSet):
    class Meta:
        model = ComplianceType
        fields = {
            "name": ["exact", "icontains"],
        }


class ComplianceTaskGroupFilter(wb_filters.FilterSet):
    class Meta:
        model = ComplianceTaskGroup
        fields = {
            "name": ["exact", "icontains"],
        }


class ComplianceTaskReviewFilter(wb_filters.FilterSet):
    class Meta:
        model = ComplianceTask
        fields = {
            "title": ["exact", "icontains"],
            "occurrence": ["exact"],
            "active": ["exact"],
            "type": ["exact"],
            "group": ["exact"],
            "risk_level": ["exact"],
        }


class ComplianceTaskFilter(ComplianceTaskReviewFilter):
    is_recurring = wb_filters.BooleanFilter(label=_("Is Recurring"), initial=True, method="filter_occurrence")
    active = wb_filters.BooleanFilter(label=_("Active"), initial=True)

    def filter_occurrence(self, queryset, name, value):
        if value is not None:
            if value:
                return queryset.filter(~Q(occurrence=ComplianceTask.Occurrence.NEVER))
            else:
                return queryset.filter(occurrence=ComplianceTask.Occurrence.NEVER)
        return queryset


class ComplianceTaskInstanceFilter(wb_filters.FilterSet):
    type_name = wb_filters.CharFilter(label=_("Administrator"), lookup_expr="icontains")
    group_name = wb_filters.CharFilter(label=_("Group"), lookup_expr="icontains")

    class Meta:
        model = ComplianceTaskInstance
        fields = {"task": ["exact"], "occured": ["lte", "gte"], "status": ["exact"], "review": ["exact"]}


class ComplianceTaskMatrixFilter(wb_filters.FilterSet):
    type_name = wb_filters.CharFilter(label=_("Administrator"), lookup_expr="icontains")
    group_name = wb_filters.CharFilter(label=_("Group"), lookup_expr="icontains")
    task_title = wb_filters.CharFilter(label=_("Tasks"), lookup_expr="icontains")

    class Meta:
        model = ComplianceTaskInstance
        fields = {}


class ComplianceActionFilter(wb_filters.FilterSet):
    active = wb_filters.BooleanFilter(label=_("Active"), initial=True)

    class Meta:
        model = ComplianceAction
        fields = {
            "title": ["exact", "icontains"],
            "deadline": ["exact"],
            "progress": ["exact"],
            "status": ["exact"],
            "active": ["exact"],
            "type": ["exact"],
            "creator": ["exact"],
            "changer": ["exact"],
            "last_modified": ["lte", "gte"],
        }


class ComplianceEventFilter(wb_filters.FilterSet):
    active = wb_filters.BooleanFilter(label=_("Active"), initial=True)

    class Meta:
        model = ComplianceEvent
        fields = {
            "title": ["exact", "icontains"],
            "type": ["exact"],
            "level": ["exact"],
            "active": ["exact"],
            "type_event": ["exact"],
            "creator": ["exact"],
            "last_modified": ["lte", "gte"],
            "confidential": ["exact"],
        }


class ReviewComplianceTaskFilter(wb_filters.FilterSet):
    is_instance = wb_filters.BooleanFilter(label=_("Is occurrence"), initial=True)

    class Meta:
        model = ReviewComplianceTask
        fields = {
            "year": ["exact"],
            "from_date": ["lte", "gte"],
            "to_date": ["lte", "gte"],
            "title": ["exact", "icontains"],
            "is_instance": ["exact"],
            "status": ["exact"],
            "changer": ["exact"],
            "changed": ["lte", "gte"],
            "occured": ["lte", "gte"],
            "review_task": ["exact"],
            "occurrence": ["exact"],
            "type": ["exact"],
        }



from datetime import date
from typing import Any, Iterable

from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils.functional import cached_property

from .checks import RiskCheck, evaluate_as_task
from .rules import RiskRule


class RiskCheckMixin(models.Model):
    """
    A utility mixin to inherit from when a model proposes a risk check workflow on one of its field
    """

    id: int

    @property
    def checked_object(self) -> Any:
        raise NotImplementedError()

    @property
    def check_evaluation_date(self) -> Any:
        raise NotImplementedError()

    @cached_property
    def checked_object_content_type(self) -> ContentType:
        return ContentType.objects.get_for_model(self.checked_object)

    @cached_property
    def activator_id(self) -> int:
        return self.id

    @cached_property
    def activator_content_type(self) -> ContentType:
        return ContentType.objects.get_for_model(self)

    @property
    def checks(self) -> models.QuerySet[RiskCheck]:
        """
        Returned the check triggered by this object (self)
        Returns: A queryset of RiskCheck
        """
        return RiskCheck.all_objects.filter(
            activator_id=self.activator_id,
            activator_content_type=self.activator_content_type,
            checked_object_content_type=self.checked_object_content_type,
            checked_object_id=self.checked_object.id,
            evaluation_date=self.check_evaluation_date,
        )

    @property
    def has_non_successful_checks(self) -> bool:
        """
        Return True if checks are available and they all succeed
        """
        return (
            self.checks.exists()
            and self.checks.exclude(status__in=[RiskCheck.CheckStatus.SUCCESS, RiskCheck.CheckStatus.WARNING]).exists()
        )

    @property
    def has_all_check_completed(self) -> bool:
        """
        Return True if checks are available and they all succeed
        """
        return (
            self.checks.exists()
            and not self.checks.filter(
                status__in=[RiskCheck.CheckStatus.RUNNING, RiskCheck.CheckStatus.PENDING]
            ).exists()
        )

    def get_worst_check_status(self) -> RiskCheck.CheckStatus:
        status_ordered = [
            RiskCheck.CheckStatus.FAILED,
            RiskCheck.CheckStatus.WARNING,
            RiskCheck.CheckStatus.RUNNING,
            RiskCheck.CheckStatus.PENDING,
        ]
        for status in status_ordered:
            if self.checks.filter(status=status).exists():
                return status
        return RiskCheck.CheckStatus.SUCCESS

    def evaluate_active_rules(
        self,
        evaluation_date: date,
        *dto,
        asynchronously: bool = True,
        ignore_breached_objects: Iterable[models.Model] | None = None,
    ):
        # serialize the ignore breached object list into a (id, content_type.id) list of tuple
        ignore_breached_object_content_types = [
            (o.id, ContentType.objects.get_for_model(o).id) for o in ignore_breached_objects
        ]
        for rule in RiskRule.objects.get_active_rules_for_object(self.checked_object):
            check = RiskCheck.all_objects.update_or_create(
                rule=rule,
                evaluation_date=evaluation_date,
                checked_object_content_type=self.checked_object_content_type,
                checked_object_id=self.checked_object.id,
                activator_id=self.activator_id,
                activator_content_type=self.activator_content_type,
                defaults={"status": RiskCheck.CheckStatus.PENDING},
            )[0]
            if asynchronously:
                transaction.on_commit(
                    lambda check_id=check.id: evaluate_as_task.delay(
                        check_id,
                        *dto,
                        override_incident=True,
                        ignore_informational_threshold=True,
                        ignore_breached_object_content_types=ignore_breached_object_content_types,
                    )
                )
            else:
                check.evaluate(
                    *dto,
                    override_incident=True,
                    ignore_informational_threshold=True,
                    ignore_breached_objects=ignore_breached_objects,
                )

    class Meta:
        abstract = True



from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Group
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    Count,
    F,
    OuterRef,
    QuerySet,
    Subquery,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from rest_framework.reverse import reverse
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.markdown.utils import custom_url_fetcher
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from weasyprint import HTML

from .compliance_type import ComplianceDocumentMixin, ComplianceType, can_active_request


class ComplianceFormType(WBModel):
    class Type(models.TextChoices):
        TEXT = "TEXT", "Text"
        FORM = "FORM", "Form"

    class Meta:
        verbose_name = "Compliance Form Type"
        verbose_name_plural = "Compliance Form Types"

    name = models.CharField(max_length=255, verbose_name=_("Name"))
    type = models.CharField(
        max_length=32,
        default=Type.TEXT,
        choices=Type.choices,
        verbose_name=_("Type"),
    )

    def __str__(self) -> str:
        return "{}".format(self.name)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceformtype"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:complianceformtyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class ComplianceForm(ComplianceDocumentMixin, WBModel):
    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        ACTIVATION_REQUESTED = "ACTIVATION_REQUESTED", "Activation Requested"
        ACTIVE = "ACTIVE", "Active"

    class Meta:
        verbose_name = "Compliance Form"
        verbose_name_plural = "Compliance Forms"

        notification_types = [
            create_notification_type(
                code="wbcompliance.complianceform.notify",
                title="Compliance Form Notification",
                help_text="Sends out a notification when something happens with a compliance form",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        verbose_name=_("Creator"),
        related_name="compliance_forms",
        on_delete=models.deletion.SET_NULL,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        "directory.Person", null=True, blank=True, verbose_name=_("Changer"), on_delete=models.deletion.SET_NULL
    )
    changed = models.DateTimeField(auto_now=True, verbose_name=_("Changed"))
    title = models.CharField(max_length=255, verbose_name=_("Title"))
    policy = models.TextField(default="", null=True, blank=True, verbose_name=_("Policy"))
    start = models.DateField(verbose_name=_("Start"))
    end = models.DateField(verbose_name=_("End"), null=True, blank=True)
    assigned_to = models.ManyToManyField(
        Group,
        related_name="forms_of_group",
        blank=True,
        verbose_name=_("Group to which the Form applies"),
    )
    only_internal = models.BooleanField(
        default=True, verbose_name=_("Only internal"), help_text=_("Send the Form only to internal users")
    )
    form_type = models.ForeignKey(to=ComplianceFormType, on_delete=models.PROTECT, verbose_name=_("Form Type"))

    compliance_type = models.ForeignKey(to=ComplianceType, on_delete=models.PROTECT, verbose_name=_("Type"))

    status = FSMField(
        default=Status.DRAFT,
        choices=Status.choices,
        verbose_name=_("Status"),
        help_text=_("The Compliance Form status (default to Draft)"),
    )
    version = models.IntegerField(default=0)

    def generate_pdf(self) -> bytes:
        html_content = ""
        if self.form_type.type == ComplianceFormType.Type.TEXT:
            html_content = self.policy
        elif self.form_type.type == ComplianceFormType.Type.FORM:
            html = get_template("compliance/compliance_form.html")
            html_content = html.render(
                {"form_type": self.form_type.type, "today": timezone.now(), "form": self, "is_signature": False}
            )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.ACTIVATION_REQUESTED,
        permission=lambda instance, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:complianceform",),
                icon=WBIcon.EDIT.icon,
                key="activationrequested",
                label=_("Request Activation"),
                action_label=_("Request Activation"),
                description_fields=_(
                    "{{_form_type.name}} <p>Title: <b>{{title}}</b></p>\
                <p>Version: <b>{{version}}</b></p> <p>Status: <b>{{status}}</b></p> <p>Start: <b>{{start}}</b></p>\
                <p>End: <b>{{end}}</b></p> <p>Do you want to send this request for validation ?</p>"
                ),
            )
        },
    )
    def activationrequested(self, by=None, description=None, **kwargs):
        if self.compliance_type:
            # notify the compliance team without the current user
            if by:
                self.changer = by.profile
            current_profile = self.changer if self.changer else self.creator
            recipients = ComplianceType.get_administrators(self.compliance_type).exclude(profile=current_profile)
            msg = _("Validation Request from {} to activate a {} version <b>{}</b> : <b>{}</b>").format(
                str(current_profile), self.form_type, self.version, self.title
            )
            title = _("Validation Request {} : {}").format(self.form_type, self.title)
            self.notify(title, msg, recipients, admin_compliance=True)

    @transition(
        field=status,
        source=Status.ACTIVATION_REQUESTED,
        target=Status.ACTIVE,
        permission=can_active_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:complianceform",),
                icon=WBIcon.SEND.icon,
                key="active",
                label=_("Activate"),
                action_label=_("Activate"),
                description_fields=_(
                    """
                <p> Title: <b> {{title}} </b></p> <p>Version: <b>{{version}}</b></p>
                <p>Start: <b>{{start}}</b></p><p>End: <b>{{end}}</b></p>
                <p>Do you want to activate this {{_form_type.name}} ?</p>
                """
                ),
            )
        },
    )
    def active(self, by=None, description=None, **kwargs):
        self.version += 1
        if by:
            self.changer = by.profile
        current_profile = self.changer if self.changer else self.creator
        self.create_compliance_form_signature()

        msg = _("<p>{} has activated a {} version <b>{}</b> : <b>{}</b></p>").format(
            str(current_profile), self.form_type, self.version, self.title
        )
        if self.policy and self.policy != "<p></p>" and self.policy != "null":
            msg += _("</br><p><b>Policy:</b></p><i>{}</i>").format(self.policy)
        title = _("Applying a {} : {}").format(self.form_type, self.title)
        self.notify(title, msg)
        self.save()

    @transition(
        field=status,
        source=[Status.ACTIVATION_REQUESTED, Status.ACTIVE],
        target=Status.DRAFT,
        permission=lambda instance, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:complianceform",),
                icon=WBIcon.EDIT.icon,
                key="draft",
                label=_("Return to Draft Mode"),
                action_label=_("Return to Draft Mode"),
                description_fields=_(
                    """
                {{_form_type.name}}<p>Title: <b> {{title}} </b></p><p>Version: <b>{{version}}</b></p>
                <p>Status: <b>{{status}}</b></p> <p>Start: <b>{{start}}</b></p>  <p>End: <b>{{end}}</b></p>
                <p>Do you want to return to draft mode this {{_form_type.name}}?</p>
                """
                ),
            )
        },
    )
    def draft(self, by=None, description=None, **kwargs):
        if self.compliance_type:
            if by:
                self.changer = by.profile
            current_profile = self.changer if self.changer else self.creator
            if self.status == ComplianceForm.Status.ACTIVE:
                msg = _("{} has disabled a {} version <b>{}</b> : <b>{}</b>").format(
                    str(current_profile), self.form_type, self.version, self.title
                )
                title = _("Disabled {} : {}").format(self.form_type, self.title)
            else:
                msg = _("{} has modified a {} version <b>{}</b> to Draft : <b>{}</b>").format(
                    str(current_profile), self.form_type, self.version, self.title
                )
                title = _("{} Drafted: {}").format(self.form_type, self.title)
            recipients = ComplianceType.get_administrators(self.compliance_type).exclude(profile=current_profile)
            self.notify(title, msg, recipients, admin_compliance=True)

    def get_signing_users(self) -> QuerySet["User"]:
        users = User.objects.filter(is_active=True)
        if self.only_internal:
            users = users.filter_internal()
        if self.assigned_to.exists():
            users = users.filter(groups__in=self.assigned_to.all())
        return users.distinct()

    def create_compliance_form_signature(self) -> None:
        for user in self.get_signing_users():
            compliance_form_signature = ComplianceFormSignature.objects.create(
                compliance_form=self,
                version=self.version,
                policy=self.policy,
                start=self.start,
                end=self.end,
                person=user.profile,
            )
            sections = ComplianceFormSection.objects.filter(compliance_form=compliance_form_signature.compliance_form)
            for section in sections:
                signature_section = ComplianceFormSignatureSection.objects.create(
                    compliance_form_signature=compliance_form_signature, name=section.name
                )
                rules = ComplianceFormRule.objects.filter(section=section)
                for rule in rules:
                    ComplianceFormSignatureRule.objects.create(section=signature_section, text=rule.text)

    def notify(
        self, title: str, msg: str, recipients: QuerySet["User"] | None = None, admin_compliance: bool = False
    ) -> None:
        """
        param:
            recipients: list of users, if not set, we get all users assigned to the form
            admin_compliance: by default is false, allow to send the signature form to recipient, if true send rather the form to admin.
        """
        if admin_compliance:
            users = ComplianceType.get_administrators(self.compliance_type)
            if recipients:
                users = users.filter(id__in=[_user.id for _user in recipients])
        else:
            users = recipients if recipients else self.get_signing_users()
        for user in users:
            if admin_compliance:
                endpoint = reverse("wbcompliance:complianceform-detail", args=[self.id])
            else:
                endpoint = None
                if formsignature := (
                    ComplianceFormSignature.objects.filter(compliance_form=self, person=user.profile)
                    .order_by("version")
                    .last()
                ):
                    endpoint = reverse("wbcompliance:complianceformsignature-detail", args=[formsignature.id])

            if endpoint:
                send_notification(
                    code="wbcompliance.complianceform.notify",
                    title=title,
                    body=msg,
                    user=user,
                    endpoint=endpoint,
                )

    @classmethod
    def get_subquery_total_compliance_form_signature(cls, remaining_signed: bool = False) -> Subquery:
        if remaining_signed:
            compliance_form_signatures = ComplianceFormSignature.objects.filter(
                compliance_form=OuterRef("pk"),
                signed=None,
            ).order_by("-version")
        else:
            compliance_form_signatures = ComplianceFormSignature.objects.filter(
                compliance_form=OuterRef("pk"),
            ).order_by("-version")

        return Coalesce(
            Subquery(
                compliance_form_signatures.values("compliance_form__pk")
                .annotate(total_signed=Count("compliance_form__pk"))
                .values("total_signed")[:1]
            ),
            0,
        )

    @classmethod
    def get_subquery_compliance_form_signature(cls, person_signed) -> Subquery:
        compliance_form_signatures = ComplianceFormSignature.objects.filter(
            compliance_form=OuterRef("pk"), person=person_signed
        ).order_by("-version")
        return Coalesce(
            Subquery(
                compliance_form_signatures.values("compliance_form__pk")
                .annotate(
                    is_signed=Case(
                        When(signed=None, then=Value(False)),
                        default=Value(True),
                        output_field=BooleanField(),
                    )
                )
                .values("is_signed")[:1]
            ),
            None,
        )

    def __str__(self) -> str:
        return "{} {} ({} - {}) ".format(self.title, self.version, self.start, self.end)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceform"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:complianceformrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ComplianceFormSignature(ComplianceDocumentMixin, models.Model):
    class Meta:
        verbose_name = "Compliance Form Signature"
        verbose_name_plural = "Compliance Signatures"

        notification_types = [
            create_notification_type(
                code="wbcompliance.complianceformsignature.signed",
                title="Compliance Form Signature Confirmation",
                help_text="Send a notification as a confirmation that a compliance form has been signed",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    compliance_form = models.ForeignKey(
        to="wbcompliance.ComplianceForm",
        verbose_name=_("Form"),
        related_name="complianceforms",
        on_delete=models.CASCADE,
    )
    version = models.IntegerField(default=0, verbose_name=_("Version"))
    start = models.DateField(verbose_name=_("Start"))
    end = models.DateField(verbose_name=_("End"), null=True, blank=True)
    policy = models.TextField(default="", null=True, blank=True, verbose_name=_("Policy"))
    signed = models.DateTimeField(null=True, blank=True, verbose_name=_("Signed"))
    person = models.ForeignKey(
        to="directory.Person",
        verbose_name=_("Signer"),
        related_name="signed_compliance_forms",
        on_delete=models.CASCADE,
    )
    remark = models.TextField(null=True, blank=True, default="", verbose_name=_("Remark"))

    def __str__(self) -> str:
        return "{} {} ({} - {})".format(self.compliance_form.title, self.version, self.start, self.end)

    def generate_pdf(self) -> bytes:
        html_content = ""
        html = get_template("compliance/compliance_form.html")
        html_content = html.render(
            {
                "form_type": self.compliance_form.form_type.type,
                "today": timezone.now(),
                "form": self,
                "is_signature": True,
            }
        )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceformsignature"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{compliance_form.title}} {{self.version}} ({{start}} - {{end}})"


class ComplianceFormSection(WBModel):
    """Model that represents a section of the Compliance Form"""

    compliance_form = models.ForeignKey(
        ComplianceForm, related_name="compliance_forms", verbose_name=_("Compliance Form"), on_delete=models.CASCADE
    )
    name = models.CharField(max_length=255, verbose_name=_("Name section"))

    class Meta:
        verbose_name = "Section of the Compliance Form"
        verbose_name_plural = "Sections of the Compliance Form"

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceformsection"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:complianceformsectionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class ComplianceFormRule(models.Model):
    """Model that represents a rule in a Section of the Compliance Form"""

    section = models.ForeignKey(
        ComplianceFormSection, related_name="rules", verbose_name=_("Section"), on_delete=models.CASCADE
    )
    text = models.TextField(default="")
    ticked = models.BooleanField(
        default=False,
        verbose_name=_("Expected Answer"),
    )

    class Meta:
        verbose_name = "Rule of the section of the Compliance Form"
        verbose_name_plural = "Rules of the section of the Compliance Form"

    def __str__(self) -> str:
        return "{} {}".format(self.section, self.id)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceformrule"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{section}} {{id}}"


class ComplianceFormSignatureSection(models.Model):
    """Model that represents a section of the Compliance Form Signature"""

    compliance_form_signature = models.ForeignKey(
        ComplianceFormSignature,
        related_name="compliance_form_signatures",
        verbose_name=_("Compliance Form Signature"),
        on_delete=models.CASCADE,
    )
    name = models.CharField(max_length=255, verbose_name=_("Name section"))

    class Meta:
        verbose_name = "Section of the Compliance Form Signature"
        verbose_name_plural = "Sections of the Compliance Form Signature"

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:complianceformsignaturesectionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class ComplianceFormSignatureRule(models.Model):
    """Model that represents a rule in a Section of the Compliance Form Signature"""

    section = models.ForeignKey(
        ComplianceFormSignatureSection, related_name="rules", verbose_name=_("Section"), on_delete=models.CASCADE
    )
    text = models.TextField(default="")
    ticked = models.BooleanField(
        default=False,
        verbose_name=_("Answer"),
    )
    comments = models.TextField(default="")

    class Meta:
        verbose_name = "Rule of the section of the Compliance Form Signature"
        verbose_name_plural = "Rules of the section of the Compliance Form Signature"

    def __str__(self) -> str:
        return "{} {}".format(self.section, self.id)

    @classmethod
    def get_subquery_expected_ticked(cls) -> Subquery:
        return Coalesce(
            Subquery(
                ComplianceFormRule.objects.filter(
                    section__compliance_form=OuterRef("section__compliance_form_signature__compliance_form"),
                    section__name=OuterRef("section__name"),
                    text=OuterRef("text"),
                )
                .values("ticked")
                .annotate(exp=F("ticked"))
                .values("exp")[:1]
            ),
            None,
        )

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{section}} {{id}}"


if apps.is_installed("wbhuman_resources"):

    @receiver(post_save, sender="wbhuman_resources.EmployeeHumanResource")
    def post_save_compliance_form_employee(sender, instance, created, **kwargs):
        if created:
            for compliance_form in ComplianceForm.objects.filter(status=ComplianceForm.Status.ACTIVE):
                if compliance_form.get_signing_users().filter(profile_id=instance.profile.id).exists():
                    compliance_form_signature, created = ComplianceFormSignature.objects.get_or_create(
                        compliance_form=compliance_form,
                        version=compliance_form.version,
                        policy=compliance_form.policy,
                        start=compliance_form.start,
                        end=compliance_form.end,
                        person=instance.profile,
                    )
                    sections = ComplianceFormSection.objects.filter(
                        compliance_form=compliance_form_signature.compliance_form
                    )
                    for section in sections:
                        signature_section = ComplianceFormSignatureSection.objects.create(
                            compliance_form_signature=compliance_form_signature, name=section.name
                        )
                        rules = ComplianceFormRule.objects.filter(section=section)
                        for rule in rules:
                            ComplianceFormSignatureRule.objects.create(section=signature_section, text=rule.text)
                    if _user := getattr(instance.profile, "user_account", None):
                        msg = _("<p>{} has activated a <b>{}</b> policy version <b>{} </b></p>").format(
                            str(compliance_form.changer), compliance_form.title, compliance_form.version
                        )
                        if (
                            compliance_form.policy
                            and compliance_form.policy != "<p></p>"
                            and compliance_form.policy != "null"
                        ):
                            msg += _("</br><p><b>Policy:</b></p><i>{}</i>").format(compliance_form.policy)
                        title = _("Applying a {} : {}").format(compliance_form.form_type, compliance_form.title)
                        compliance_form.notify(title, msg, recipients=[_user])



import pytest
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import User

from wbcompliance.models import ComplianceForm, ComplianceFormSignature

from .mixins import UserTestMixin


@pytest.mark.django_db
class TestSpecificModels(UserTestMixin):
    @pytest.mark.parametrize("other_user", [True, False])
    def test_draft_activationrequested_active(self, compliance_form_factory, other_user):
        other_user = UserFactory() if other_user else None
        compliance_form = compliance_form_factory()
        assert compliance_form.status == ComplianceForm.Status.DRAFT
        compliance_form.activationrequested(by=other_user)
        assert compliance_form.status == ComplianceForm.Status.ACTIVATION_REQUESTED
        compliance_form.active(by=other_user)
        assert compliance_form.status == ComplianceForm.Status.ACTIVE
        compliance_form.draft(by=other_user)
        assert compliance_form.status == ComplianceForm.Status.DRAFT

    def test_create_compliance_form_signature(
        self,
        compliance_form_factory,
        user,
        compliance_form_section_factory,
        compliance_form_rule_factory,
    ):
        compliance_form = compliance_form_factory(
            assigned_to=(user.groups.filter(name="Compliance Position").first(),)
        )
        section = compliance_form_section_factory(compliance_form=compliance_form)
        compliance_form_rule_factory(section=section)

        compliance_form.activationrequested()
        compliance_form.active()
        assert ComplianceFormSignature.objects.count() == 1

        # test empty assigned_to
        compliance_form1 = compliance_form_factory(only_internal=True)
        compliance_form1.activationrequested()
        compliance_form1.active()
        assert (
            ComplianceFormSignature.objects.filter(compliance_form=compliance_form1).count()
            == User.objects.filter_internal().count()
        )

    # def test_only_internal(self, compliance_form_factory):
    #     employee.contract_type = EmployeeHumanResource.ContractType.EXTERNAL
    #     employee.save()
    #     compliance_form = compliance_form_factory(assigned_to=(user.groups.filter(name="Compliance Position").first(),), only_internal=True)
    #     assert compliance_form.get_signing_users().count() == 0
    #
    #     employee2 = EmployeeHumanResourceFactory(profile=person2)
    #     compliance_form2 = compliance_form_factory(assigned_to=(employee2.position,), only_internal=False)
    #     assert compliance_form2.get_signing_users().count() == 1



import random
from datetime import date, timedelta
from unittest.mock import patch

import pytest
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from faker import Faker
from guardian.utils import get_anonymous_user
from psycopg.types.range import NumericRange
from wbcore.contrib.permission.utils import perm_to_permission

from wbcompliance.models import (
    RiskIncident,
    RiskRule,
    RuleCheckedObjectRelationship,
    RuleThreshold,
)
from wbcompliance.models.risk_management.rules import process_rule_as_task

fake = Faker()


@pytest.mark.django_db
class TestRuleCheckedObjectRelationship:
    def test_init(self, rule_checked_object_relationship):
        assert rule_checked_object_relationship

    def test_init_wrong_content_type(self, risk_rule, rule_checked_object_relationship_factory):
        risk_rule.rule_backend.allowed_checked_object_content_type = ContentType.objects.get_for_model(RuleThreshold)
        risk_rule.rule_backend.save()
        with pytest.raises(ValidationError):
            rule_checked_object_relationship_factory.create(
                rule=risk_rule,
                checked_object_content_type=ContentType.objects.get_for_model(RuleCheckedObjectRelationship),
            )


@pytest.mark.django_db
class TestRuleBackend:
    def test_init(self, rule_backend):
        assert rule_backend

    def test_backend_class(self, rule_backend):
        assert rule_backend.backend_class.__module__
        assert rule_backend.backend_class.__class__

    @pytest.mark.parametrize("evaluation_date", [fake.date_object()])
    def test_backend(self, rule_backend, risk_rule, evaluation_date):
        assert rule_backend.backend(evaluation_date, risk_rule, risk_rule.parameters, RuleThreshold.objects.none())


@pytest.mark.django_db
class TestRuleThreshold:
    def test_init(self, rule_threshold):
        assert rule_threshold

    @pytest.mark.parametrize(
        "lower,upper,value,result",
        [
            (0, 1, random.random(), True),
            (1, 2, random.random(), False),
            (-1, 0, random.random(), False),
            (None, 1, random.random(), True),
            (0, None, random.random(), True),
        ],
    )
    def test_is_inrange(self, rule_threshold_factory, lower, upper, value, result):
        threshold = rule_threshold_factory.create(range=NumericRange(lower=lower, upper=upper))
        assert threshold.is_inrange(value) == result

    def test_get_notifiable_users(self, rule_threshold, user_factory):
        assert not rule_threshold.get_notifiable_users().exists()
        user = user_factory.create()
        user_group = user_factory.create()
        group_test = Group.objects.create(name="test")
        user_group.groups.add(group_test)
        rule_threshold.notifiable_users.add(user.profile)  # Add user to threshold person group
        rule_threshold.notifiable_groups.add(group_test)  # Add user to threshold group
        users = rule_threshold.get_notifiable_users()
        assert set(users.values_list("id", flat=True)) == {user.id, user_group.id}


@pytest.mark.django_db
class TestRiskRule:
    def test_init(self, risk_rule):
        assert risk_rule

    def test_checked_objects(self, risk_rule_factory, rule_threshold_factory):
        rule = risk_rule_factory.create()
        other_rule = risk_rule_factory.create()
        threshold1 = rule_threshold_factory.create()
        threshold2 = rule_threshold_factory.create()
        RuleCheckedObjectRelationship.objects.create(checked_object=threshold1, rule=rule)
        RuleCheckedObjectRelationship.objects.create(checked_object=get_anonymous_user(), rule=rule)
        RuleCheckedObjectRelationship.objects.create(checked_object=threshold2, rule=other_rule)
        assert set(rule.checked_objects) == {threshold1, get_anonymous_user()}

    def test_checks(self, risk_rule_factory, rule_checked_object_relationship_factory, risk_check_factory):
        rule = risk_rule_factory.create()
        other_rule = risk_rule_factory.create()
        rel1 = rule_checked_object_relationship_factory.create(rule=rule)
        rel2 = rule_checked_object_relationship_factory.create(rule=rule)
        other_rel = rule_checked_object_relationship_factory.create(rule=other_rule)
        check1 = risk_check_factory.create(rule_checked_object_relationship=rel1)
        check2 = risk_check_factory.create(rule_checked_object_relationship=rel2)
        risk_check_factory.create(rule_checked_object_relationship=other_rel)
        assert set(rule.checks.values_list("id", flat=True)) == {check1.id, check2.id}

    @pytest.mark.parametrize(
        "risk_rule__automatically_close_incident, evaluation_date, expected_incident_status",
        [
            (True, fake.date_object(), RiskIncident.Status.CLOSED),
            (False, fake.date_object(), RiskIncident.Status.OPEN),
        ],
    )
    def test_process_rule(
        self,
        risk_rule,
        rule_checked_object_relationship_factory,
        rule_threshold_factory,
        evaluation_date,
        expected_incident_status,
    ):
        rule_threshold_factory.create(rule=risk_rule)
        rule_checked_object_relationship_factory.create(rule=risk_rule)

        risk_rule.process_rule(evaluation_date)
        incident = RiskIncident.objects.get(rule=risk_rule, status=expected_incident_status)
        assert incident.checked_object_relationships.count() == 1

    def test_get_permissions_for_user(self, risk_rule, user_factory):
        user_with_permission = user_factory.create()
        user_without_permission = user_factory.create()
        user_with_permission.user_permissions.add(perm_to_permission(RiskRule.view_perm_str))
        assert risk_rule.get_permissions_for_user(user_with_permission)[RiskRule.view_perm_str] is False
        assert not risk_rule.get_permissions_for_user(user_without_permission)

    def test_get_rules_for_object(self, risk_rule_factory, rule_checked_object_relationship_factory):
        rule1 = risk_rule_factory.create()
        rule2 = risk_rule_factory.create()

        rule_checked_object_relationship_factory.create(rule=rule1)
        rule_checked_object_relationship_factory.create(rule=rule2)

        res = RiskRule.get_rules_for_object(get_anonymous_user())
        assert res.exists()
        assert set(res.values_list("id", flat=True)) == {rule1.id, rule2.id}

    @patch.object(RiskRule, "process_rule")
    @pytest.mark.parametrize(
        "evaluation_date, override_incident",
        [(fake.date_object(), fake.pybool())],
    )
    def test_process_rule_as_task(self, mock_function, risk_rule, evaluation_date, override_incident):
        process_rule_as_task(risk_rule.id, evaluation_date, override_incident)
        assert mock_function.call_count == 1

    @patch("wbcompliance.models.risk_management.rules.send_notification")
    def test_do_not_notify_no_open_incidents(
        self,
        mock_fct,
        risk_incident_factory,
        rule_checked_object_relationship,
        risk_check_factory,
        checked_object_incident_relationship_factory,
    ):
        risk_rule = rule_checked_object_relationship.rule
        risk_incident = risk_incident_factory.create(rule=risk_rule)
        sub_incident = checked_object_incident_relationship_factory.create(incident=risk_incident)

        incident_date = sub_incident.incident_date
        risk_rule.notify(incident_date)
        assert mock_fct.call_count == 1

        risk_incident.refresh_from_db()
        assert risk_incident.is_notified is True

        # renotifying this incident still triggers the method because the incident (even marked as notified) is still open
        risk_rule.notify(incident_date)
        assert mock_fct.call_count == 2

        risk_incident.ignore()
        risk_incident.save()
        risk_rule.notify(incident_date)
        assert mock_fct.call_count == 2

        tomorrow = incident_date + timedelta(days=1)
        tomorrow_risk_check = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship, evaluation_date=tomorrow
        )
        tomorrow_sub_incident = checked_object_incident_relationship_factory.create(  # noqa
            rule_check=tomorrow_risk_check, incident=risk_incident
        )

        risk_incident.refresh_from_db()
        assert risk_incident.status == RiskIncident.Status.OPEN
        # the incident was already notified and is closed, so another subincident will mark the incident as open and retrigger notification
        risk_rule.notify(tomorrow)
        assert mock_fct.call_count == 3

        after_tomorrow = tomorrow + timedelta(days=1)
        risk_incident.ignore()
        risk_incident.last_ignored_date = tomorrow
        risk_incident.ignore_duration = timedelta(days=1)
        risk_incident.save()

        # this shoudn't reopen the incident
        after_tomorrow_risk_check = risk_check_factory.create(
            rule_checked_object_relationship=rule_checked_object_relationship, evaluation_date=after_tomorrow
        )
        after_tomorrow_sub_incident = checked_object_incident_relationship_factory.create(  # noqa
            rule_check=after_tomorrow_risk_check, incident=risk_incident
        )
        risk_incident.refresh_from_db()
        assert risk_incident.status == RiskIncident.Status.IGNORED
        # then notify shouldn't happen
        risk_rule.notify(after_tomorrow)
        assert mock_fct.call_count == 3

    def test_is_evaluation_date_valid(self, risk_rule):
        risk_rule.activation_date = None
        risk_rule.save()
        assert risk_rule.is_evaluation_date_valid(fake.date_object()) is False

        activation_date = fake.date_object()
        risk_rule.activation_date = activation_date
        risk_rule.save()
        assert (
            risk_rule.is_evaluation_date_valid(
                fake.date_between(activation_date, activation_date + timedelta(days=365))
            )
            is True
        )
        assert (
            risk_rule.is_evaluation_date_valid(
                fake.date_between(activation_date - timedelta(days=365), activation_date - timedelta(days=1))
            )
            is False
        )

        # enable a rule to be checked every week on monday (i.e. every monday)
        activation_date = date(2024, 11, 25)  # Monday
        risk_rule.activation_date = activation_date
        risk_rule.frequency = "RRULE:FREQ=WEEKLY"
        risk_rule.save()

        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 25)) is True
        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 26)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 27)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 28)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 29)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 11, 30)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 12, 1)) is False
        assert risk_rule.is_evaluation_date_valid(date(2024, 12, 2)) is True



from .compliance_form import (
    ComplianceForm,
    ComplianceFormRule,
    ComplianceFormSection,
    ComplianceFormSignature,
    ComplianceFormSignatureRule,
    ComplianceFormSignatureSection,
    ComplianceFormType,
)
from .compliance_task import (
    ComplianceAction,
    ComplianceEvent,
    ComplianceTask,
    ComplianceTaskGroup,
    ComplianceTaskInstance,
    ComplianceType,
    ReviewComplianceTask,
)
from .compliance_type import ComplianceType, update_or_create_compliance_document
from .risk_management import *



from .compliance_form import (
    CFComplianceFormSectionEndpointConfig,
    CFComplianceFormSignatureEndpointConfig,
    ComplianceFormEndpointConfig,
    ComplianceFormRuleEndpointConfig,
    ComplianceFormSectionRuleEndpointConfig,
    ComplianceFormSignatureEndpointConfig,
    ComplianceFormSignatureSectionRuleEndpointConfig,
    ComplianceFormTypeEndpointConfig,
)
from .compliance_task import (
    ComplianceActionEndpointConfig,
    ComplianceEventEndpointConfig,
    ComplianceTaskComplianceTaskGroupEndpointConfig,
    ComplianceTaskEndpointConfig,
    ComplianceTaskGroupEndpointConfig,
    ComplianceTaskInstanceComplianceTaskEndpointConfig,
    ComplianceTaskInstanceEndpointConfig,
    ComplianceTaskInstanceReviewGroupEndpointConfig,
    ComplianceTaskInstanceReviewNoGroupEndpointConfig,
    ComplianceTaskMatrixEndpointConfig,
    ComplianceTaskReviewGroupEndpointConfig,
    ComplianceTaskReviewNoGroupEndpointConfig,
    ReviewComplianceTaskEndpointConfig,
)
from .compliance_type import ComplianceTypeEndpointConfig
from .risk_managment import *



from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class RiskCheckButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(key="incidents", label=_("Incidents"), icon=WBIcon.WARNING.icon),
        }



from .compliance_form import (
    ComplianceFormModelSerializer,
    ComplianceFormRepresentationSerializer,
    ComplianceFormRuleModelSerializer,
    ComplianceFormSectionModelSerializer,
    ComplianceFormSectionRepresentationSerializer,
    ComplianceFormSignatureModelSerializer,
    ComplianceFormSignatureRuleModelSerializer,
    ComplianceFormSignatureSectionRepresentationSerializer,
    ComplianceFormTypeModelSerializer,
    ComplianceFormTypeRepresentationSerializer,
)
from .compliance_task import (
    ComplianceActionModelSerializer,
    ComplianceEventModelSerializer,
    ComplianceTaskGroupModelSerializer,
    ComplianceTaskGroupRepresentationSerializer,
    ComplianceTaskInstanceListModelSerializer,
    ComplianceTaskInstanceModelSerializer,
    ComplianceTaskModelSerializer,
    ComplianceTaskRepresentationSerializer,
    ComplianceTaskReviewModelSerializer,
    ReviewComplianceTaskModelSerializer,
    ReviewComplianceTaskRepresentationSerializer,
)
from .compliance_type import (
    ComplianceTypeModelSerializer,
    ComplianceTypeRepresentationSerializer,
)
from .risk_management import *



import pytest
from django.contrib.auth.models import Group
from faker import Faker
from wbcore.contrib.authentication.factories import InternalUserFactory

fake = Faker()


class UserTestMixin:
    @pytest.fixture()
    def user(self):
        user = InternalUserFactory.create()
        group = Group.objects.create(name="Compliance Position")
        user.groups.add(group)
        return user



from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.serializers import GroupRepresentationSerializer
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbcore.contrib.permission.internal.registry import get_internal_groups

from wbcompliance.models import (
    ComplianceForm,
    ComplianceFormRule,
    ComplianceFormSection,
    ComplianceFormSignature,
    ComplianceFormSignatureRule,
    ComplianceFormSignatureSection,
    ComplianceFormType,
    ComplianceType,
)

from .compliance_type import ComplianceTypeRepresentationSerializer


class ComplianceFormTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = ComplianceFormType
        fields = ("id", "name", "type")


class ComplianceFormTypeModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = ComplianceFormType
        fields = ("id", "name", "type")


# COMPLIANCE FORM
class ComplianceFormRepresentationSerializer(wb_serializers.RepresentationSerializer):
    # _detail = wb_serializers.HyperlinkField(reverse_name="wbcompliance:complianceform-detail")
    class Meta:
        model = ComplianceForm
        fields = (
            "id",
            "title",
            "form_type",
            "version",
            "status",
            "start",
            "end",
        )


class ComplianceFormModelSerializer(wb_serializers.ModelSerializer):
    creator = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.CurrentUserDefault("profile"),
        many=False,
        read_only=True,
    )
    start = wb_serializers.DateField(default=timezone.now().date(), label=_("Start"))
    _creator = PersonRepresentationSerializer(source="creator")
    _changer = PersonRepresentationSerializer(source="changer")
    current_signed = wb_serializers.CharField(
        label=_("Current Signed"),
        read_only=True,
        help_text=_("The number of signatures of the compliance form over the number of total to be signed"),
    )
    is_signed = wb_serializers.BooleanField(required=False, read_only=True)
    _assigned_to = GroupRepresentationSerializer(source="assigned_to", many=True)
    _form_type = ComplianceFormTypeRepresentationSerializer(source="form_type")
    compliance_type = wb_serializers.PrimaryKeyRelatedField(
        label=_("Administrator"), queryset=ComplianceType.objects.all()
    )
    _compliance_type = ComplianceTypeRepresentationSerializer(source="compliance_type")

    @cached_property
    def user_profile(self) -> Person | None:
        if request := self.context.get("request"):
            return request.user.profile
        return None

    class Meta:
        model = ComplianceForm
        fields = (
            "id",
            "creator",
            "_creator",
            "created",
            "changer",
            "_changer",
            "changed",
            "title",
            "policy",
            "status",
            "version",
            "_additional_resources",
            "assigned_to",
            "_assigned_to",
            "is_signed",
            "current_signed",
            "only_internal",
            "start",
            "end",
            "form_type",
            "_form_type",
            "compliance_type",
            "_compliance_type",
        )

        read_only_fields = ("creator", "created", "changer", "changed", "version")

    def create(self, validated_data):
        validated_data["creator"] = self.user_profile
        if len(validated_data["assigned_to"]) == 0:
            validated_data["assigned_to"] = get_internal_groups()
        return super().create(validated_data)

    def update(self, instance, validated_data):
        profile = self.user_profile
        validated_data["changer"] = profile
        validated_data.pop("creator", None)
        validated_data.pop("created", None)
        persons = validated_data.pop("assigned_to", None)

        instance = super().update(instance, validated_data)

        # now update manytomany field assigned_to
        if persons is not None:
            instance.assigned_to.clear()
            for person in persons:
                instance.assigned_to.add(person)
            instance.save()

        return instance

    def validate(self, data):
        obj = self.instance
        start = data.get("start", obj.start if obj else None)
        end = data.get("end", obj.end if obj else None)
        title = data.get("title", obj.start if obj else None)
        errors = {}
        if not title:
            errors["title"] = ["This field is required."]
        if start and end:
            if start > end:
                errors["start"] = ["end date cannot be before start date"]
                errors["end"] = ["end date cannot be before start date"]

        if len(errors.keys()) > 0:
            raise serializers.ValidationError(errors)

        return data

    @wb_serializers.register_resource()
    def list_additional_resources(self, instance, request, user):
        return {
            "signatures": reverse(
                "wbcompliance:complianceform-signatures-list",
                args=[instance.id],
                request=request,
            ),
            "sections": reverse(
                "wbcompliance:complianceform-sections-list",
                args=[instance.id],
                request=request,
            ),
        }

    @wb_serializers.register_only_instance_resource()
    def instance_additional_resources(self, instance, request, user, view, **kwargs):
        additional_resources = dict()
        for section in ComplianceFormSection.objects.filter(compliance_form=instance.id).order_by("id"):
            additional_resources[f"rules{section.id}"] = reverse(
                "wbcompliance:complianceformsection-rules-list",
                args=[section.id],
                request=request,
            )

        if (
            view.user_has_compliance_admin_permission
            and instance.status == ComplianceForm.Status.ACTIVE
            and hasattr(instance, "is_signed")
        ):
            if not instance.is_signed:
                additional_resources["send_compliance_form_notification"] = reverse(
                    "wbcompliance:complianceform-sendcomplianceformnotification",
                    args=[instance.id],
                    request=request,
                )
        if view.user_has_compliance_admin_permission:
            additional_resources["regenerate_document"] = reverse(
                "wbcompliance:complianceform-regenerate-document", args=[instance.id], request=request
            )
        return additional_resources


# COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureModelSerializer(wb_serializers.ModelSerializer):
    compliance_form = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _compliance_form = ComplianceFormRepresentationSerializer(source="compliance_form")
    person = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _person = PersonRepresentationSerializer(source="person")
    is_signed = wb_serializers.BooleanField(required=False, read_only=True)

    class Meta:
        model = ComplianceFormSignature
        fields = (
            "id",
            "compliance_form",
            "_compliance_form",
            "version",
            "start",
            "end",
            "policy",
            "signed",
            "person",
            "_person",
            "remark",
            "is_signed",
            "_additional_resources",
        )

        read_only_fields = ("compliance_form", "version", "policy", "person", "signed", "start", "end")

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        additional_resources = dict()
        if not instance.signed and instance.person == user.profile:
            additional_resources["signature"] = reverse(
                "wbcompliance:complianceformsignature-signature",
                args=[instance.id],
                request=request,
            )
        for section in ComplianceFormSignatureSection.objects.filter(compliance_form_signature=instance.id).order_by(
            "id"
        ):
            additional_resources[f"rules{section.id}"] = reverse(
                "wbcompliance:complianceformsignaturesection-rules-list",
                args=[section.id],
                request=request,
            )
        if user.has_perm("wbcompliance.administrate_compliance") or user.profile == instance.person:
            additional_resources["regenerate_document"] = reverse(
                "wbcompliance:complianceformsignature-regenerate-document", args=[instance.id], request=request
            )
        return additional_resources


# SECION OF THE COMPLIANCE FORM
class ComplianceFormSectionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = ComplianceFormSection
        fields = ("id", "name")


class ComplianceFormSectionModelSerializer(wb_serializers.ModelSerializer):
    compliance_form = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceForm.objects.all(), read_only=lambda view: not view.new_mode
    )
    _compliance_form = ComplianceFormRepresentationSerializer(many=False, source="compliance_form")

    class Meta:
        model = ComplianceFormSection
        fields = ("id", "name", "compliance_form", "_compliance_form", "_additional_resources")

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        additional_resources = dict()
        additional_resources["rules"] = reverse(
            "wbcompliance:complianceformsection-rules-list",
            args=[instance.id],
            request=request,
        )
        return additional_resources


# RULES OF THE SECTION OF THE COMPLIANCE
class ComplianceFormRuleModelSerializer(wb_serializers.ModelSerializer):
    section = wb_serializers.PrimaryKeyRelatedField(
        queryset=ComplianceFormSection.objects.all(), read_only=lambda view: not view.new_mode
    )
    _section = ComplianceFormSectionRepresentationSerializer(many=False, source="section")
    text = wb_serializers.TextField(allow_null=False)

    class Meta:
        model = ComplianceFormRule
        fields = ("id", "text", "ticked", "section", "_section")


# REPRESENTATION SECTION OF THE COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureSectionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = ComplianceFormSignatureSection
        fields = ("id", "name")


# RULES OF THE SECTION OF THE COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureRuleModelSerializer(wb_serializers.ModelSerializer):
    _section = ComplianceFormSignatureSectionRepresentationSerializer(many=False, source="section")
    comments = wb_serializers.TextAreaField(label=_("Comments"), allow_blank=True, allow_null=True)

    expected_result = wb_serializers.BooleanField(read_only=True)
    same_answer = wb_serializers.BooleanField(read_only=True)

    class Meta:
        model = ComplianceFormSignatureRule
        fields = ("id", "text", "ticked", "comments", "section", "_section", "expected_result", "same_answer")

        read_only_fields = ("section", "text")



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbcompliance.models.risk_management import RiskIncident


class RiskIncidentButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if not self.view.kwargs.get("pk", None) and RiskIncident.can_manage(self.request.user):

            class RiskRuleTypeSerializer(wb_serializers.Serializer):
                resolve_status = wb_serializers.ChoiceField(
                    label=_("Rule Type"),
                    choices=[
                        (RiskIncident.Status.RESOLVED.name, RiskIncident.Status.RESOLVED.label),
                        (RiskIncident.Status.IGNORED.name, RiskIncident.Status.IGNORED.label),
                    ],
                    default=RiskIncident.Status.RESOLVED,
                )
                comment = wb_serializers.TextField(default="", label=_("Comment"))

            endpoint = reverse("wbcompliance:riskincident-resolveallincidents", args=[], request=self.request)
            queryset = RiskIncident.objects.all()
            if rule_id := self.view.kwargs.get("rule_id", None):
                queryset = queryset.filter(rule=rule_id)
                endpoint += f"?rule_id={rule_id}"
            return {
                bt.ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("wbcompliance:riskincident",),
                    endpoint=endpoint,
                    label=_("Resolve all open Incident"),
                    description_fields=_(
                        "<p>Are you sure you want to close the current <b>{}</b> opened incidents?</p>"
                    ).format(queryset.count()),
                    serializer=RiskRuleTypeSerializer,
                    action_label=_("resolveallincidents"),
                    title=_("Resolve all open Incident"),
                    instance_display=create_simple_display([["resolve_status"], ["comment"]]),
                )
            }
        return set()

    def get_custom_list_instance_buttons(self):
        return set()



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbcompliance.models import ComplianceForm, ComplianceFormSection, ComplianceType


# Compliance Form
class ComplianceFormEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if (
                not ComplianceType.is_administrator(self.request.user)
                or obj.status == ComplianceForm.Status.ACTIVATION_REQUESTED
            ):
                return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_create_endpoint()
        return None

    def get_delete_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if (
                not ComplianceType.is_administrator(self.request.user)
                or obj.status == ComplianceForm.Status.ACTIVATION_REQUESTED
            ):
                return None
        return super().get_delete_endpoint()


class ComplianceFormSignatureEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if obj.person != self.request.user.profile or obj.signed:
                return None
        return reverse(f"{self.view.get_model().get_endpoint_basename()}-list", request=self.request)


class CFComplianceFormSignatureEndpointConfig(ComplianceFormSignatureEndpointConfig):
    pass


# SECTION OF THE COMPLIANCE FORM
class CFComplianceFormSectionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcompliance:complianceform-sections-list",
            args=[self.view.kwargs["compliance_form_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        if self.instance and not ComplianceType.is_administrator(self.request.user):
            return None
        if self.instance and "compliance_form_id" in self.view.kwargs:
            obj = ComplianceForm.objects.get(id=self.view.kwargs.get("compliance_form_id"))
            if obj.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if "compliance_form_id" in self.view.kwargs:
            obj = ComplianceForm.objects.get(id=self.view.kwargs.get("compliance_form_id"))
            if obj.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        return super().get_create_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if "compliance_form_id" in self.view.kwargs:
            obj = ComplianceForm.objects.get(id=self.view.kwargs.get("compliance_form_id"))
            if obj.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        if "pk" in self.view.kwargs:
            return f'{self.get_endpoint()}{self.view.kwargs["pk"]}/'
        return super().get_delete_endpoint()


# RULES OF THE SECTION
class ComplianceFormRuleEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.instance and not ComplianceType.is_administrator(self.request.user):
            return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_create_endpoint()
        return None

    def get_delete_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_delete_endpoint()
        return None


class ComplianceFormSectionRuleEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcompliance:complianceformsection-rules-list",
            args=[self.view.kwargs["section_id"]],
            request=self.request,
        )

    def get_instance_endpoint(self, **kwargs):
        if self.instance and not ComplianceType.is_administrator(self.request.user):
            return None
        if self.instance and "section_id" in self.view.kwargs:
            obj = ComplianceFormSection.objects.get(id=self.view.kwargs.get("section_id"))
            if obj.compliance_form.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if "section_id" in self.view.kwargs:
            obj = ComplianceFormSection.objects.get(id=self.view.kwargs.get("section_id"))
            if obj.compliance_form.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        return super().get_create_endpoint()

    def get_delete_endpoint(self, **kwargs):
        if not ComplianceType.is_administrator(self.request.user):
            return None
        if "section_id" in self.view.kwargs:
            obj = ComplianceFormSection.objects.get(id=self.view.kwargs.get("section_id"))
            if obj.compliance_form.status == ComplianceForm.Status.ACTIVATION_REQUESTED:
                return None
        if "pk" in self.view.kwargs:
            return f'{self.get_endpoint()}{self.view.kwargs["pk"]}/'
        return super().get_delete_endpoint()


class ComplianceFormSignatureSectionRuleEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        if self.instance:
            obj = self.view.get_object()
            if (
                obj.section.compliance_form_signature.person != self.request.user.profile
                or obj.section.compliance_form_signature.signed
            ):
                return None
        return reverse(
            "wbcompliance:complianceformsignaturesection-rules-list",
            args=[self.view.kwargs["section_id"]],
            request=self.request,
        )


class ComplianceFormTypeEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        if self.instance and not ComplianceType.is_administrator(self.request.user):
            return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_create_endpoint()
        return None

    def get_delete_endpoint(self, **kwargs):
        if ComplianceType.is_administrator(self.request.user):
            return super().get_delete_endpoint()
        return None



from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.contrib.content_type.serializers import (
    ContentTypeRepresentationSerializer,
    DynamicObjectIDRepresentationSerializer,
)

from wbcompliance.models.risk_management import (
    RiskRule,
    RuleBackend,
    RuleCheckedObjectRelationship,
    RuleThreshold,
)
from wbcompliance.models.risk_management.rules import RuleGroup


class RuleGroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RuleGroup
        fields = ("id", "name")


class RuleCheckedObjectRelationshipRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RuleCheckedObjectRelationship
        fields = ("id", "computed_str", "checked_object_repr")


class RuleBackendRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RuleBackend
        fields = ("id", "name")


class RuleThresholdRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RuleThreshold
        fields = ("id", "computed_str")


class RiskRuleRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RiskRule
        fields = ("id", "name")


class GetContentTypeFromKwargs:
    requires_context = True

    def __call__(self, serializer_instance):
        if (view := serializer_instance.view) and (rule_id := view.kwargs.get("rule_id", None)):
            rule = RiskRule.objects.get(id=rule_id)
            if content_type := rule.rule_backend.allowed_checked_object_content_type:
                return content_type.id
        return None


class RuleCheckedObjectRelationshipModelSerializer(wb_serializers.ModelSerializer):
    checked_object_content_type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ContentType.objects.all(), default=GetContentTypeFromKwargs()
    )
    _checked_object_content_type = ContentTypeRepresentationSerializer(source="checked_object_content_type")
    _checked_object_id = DynamicObjectIDRepresentationSerializer(
        content_type_field_name="checked_object_content_type",
        source="checked_object_id",
        optional_get_parameters={"checked_object_content_type": "content_type"},
        depends_on=[{"field": "checked_object_content_type", "options": {}}],
    )

    class Meta:
        model = RuleCheckedObjectRelationship
        # dependency_map = {
        #     "checked_object_content_type": ["checked_object_id"],
        # }
        read_only_fields = ("computed_str", "checked_object_repr")
        fields = (
            "id",
            "rule",
            "checked_object_content_type",
            "_checked_object_content_type",
            "checked_object_id",
            "_checked_object_id",
            "checked_object_repr",
            "computed_str",
        )


class RuleBackendModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = RuleBackend
        fields = ("id", "name", "backend_class_path", "backend_class_name", "allowed_checked_object_content_type")
        read_only_fields = fields


class RiskRuleModelSerializer(wb_serializers.ModelSerializer):
    parameters__group_by = wb_serializers.CharField(read_only=True)
    _rule_backend = RuleBackendRepresentationSerializer(source="rule_backend")
    _creator = UserRepresentationSerializer(source="creator")
    parameters = wb_serializers.JSONTableField()
    open_incidents_count = wb_serializers.IntegerField(default=0, read_only=True)
    in_breach = wb_serializers.ChoiceField(
        read_only=True,
        choices=[("BREACH", "In Breach"), ("PASSED", "Passed"), ("INACTIVE", "Inactive")],
    )

    def validate(self, data):
        if (not self.instance or not self.instance.creator) and (request := self.context.get("request")):
            data["creator"] = request.user
        return super().validate(data)

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "relationships": reverse(
                "wbcompliance:riskrule-relationship-list",
                args=[instance.id],
                request=request,
            ),
            "thresholds": reverse(
                "wbcompliance:riskrule-threshold-list",
                args=[instance.id],
                request=request,
            ),
            "incidents": reverse(
                "wbcompliance:riskrule-incident-list",
                args=[instance.id],
                request=request,
            ),
        }

    class Meta:
        model = RiskRule
        read_only_fields = ("creator",)
        fields = (
            "id",
            "parameters__group_by",
            "permission_type",
            "creator",
            "_creator",
            "name",
            "description",
            "rule_backend",
            "_rule_backend",
            "is_enable",
            "only_passive_check_allowed",
            "automatically_close_incident",
            "is_silent",
            "is_mandatory",
            "apply_to_all_active_relationships",
            "parameters",
            "open_incidents_count",
            "in_breach",
            "frequency",
            "activation_date",
            "_additional_resources",
        )



from django.db import models
from wbcore.contrib.color.enums import WBColor


class IncidentSeverity(models.TextChoices):
    LOW = "LOW", "Low"
    MEDIUM = "MEDIUM", "Medium"
    HIGH = "HIGH", "High"

    @classmethod
    def get_color_map(cls) -> list:
        colors = [WBColor.GREEN_LIGHT.value, WBColor.YELLOW_LIGHT.value, WBColor.RED_LIGHT.value]
        return [choice for choice in zip(cls, colors, strict=False)]



import pytest
from wbcore.contrib.notifications.models import Notification

from wbcompliance.tasks import send_reminder_of_unsigned_compliance_form


@pytest.mark.django_db
class TestComplianceForm:
    def test_send_reminder_of_unsigned_compliance_form(
        self, compliance_form_factory, compliance_form_signature_factory
    ):
        cf1 = compliance_form_factory.create()
        cf2 = compliance_form_factory.create()
        cf3 = compliance_form_factory.create()

        compliance_form_signature_factory.create(compliance_form=cf1, signed=None)
        compliance_form_signature_factory.create(compliance_form=cf1, version=1, signed=None)
        compliance_form_signature_factory.create(compliance_form=cf2)
        compliance_form_signature_factory.create(compliance_form=cf3, signed=None)

        assert Notification.objects.count() == 0
        send_reminder_of_unsigned_compliance_form(compliance_form_ids=[cf1.id, cf2.id])
        assert Notification.objects.filter(
            title=f"Unsigned Compliance Form - {cf1.form_type} Reminder : {cf1.title}"
        ).exists()
        assert Notification.objects.filter(
            title=f"Unsigned Compliance Form - {cf2.form_type} Reminder : {cf2.title}"
        ).exists()



from django.utils.translation import gettext as _
from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbcompliance.serializers import ComplianceFormSignatureModelSerializer


class AbstractComplianceDocumentButtonConfig:
    @classmethod
    def _get_compliance_document_button(cls) -> bt.ActionButton:
        class ComplianceDocumentSerializer(wb_serializers.Serializer):
            send_email = wb_serializers.BooleanField(default=False, label=_("Email me the document"))

        return bt.ActionButton(
            method=RequestType.PATCH,
            key="regenerate_document",
            action_label=_("PDF Document is being generated"),
            title=_("Generate PDF Document"),
            label=_("Generate PDF Document"),
            icon=WBIcon.DOCUMENT.icon,
            serializer=ComplianceDocumentSerializer,
            instance_display=create_simple_display([["send_email"]]),
        )


# COMPLIANCE FORM
class ComplianceFormButtonConfig(ButtonViewConfig):
    CUSTOM_INSTANCE_BUTTONS = CUSTOM_LIST_INSTANCE_BUTTONS = {
        bt.ActionButton(
            method=RequestType.PATCH,
            identifiers=("wbcompliance:compliance_form",),
            key="send_compliance_form_notification",
            action_label=_("Send signature reminder notification"),
            title=_("Send signature reminder notification"),
            label=_("Send signature reminder notification"),
            icon=WBIcon.FEEDBACK.icon,
            description_fields=_(
                """<p> <b>{{_form_type.name}} :</b> {{title}}</p>
             <p>Do you want to send a Compliance Form reminder notification ? </p>"""
            ),
            confirm_config=bt.ButtonConfig(label=_("Confirm")),
            cancel_config=bt.ButtonConfig(label=_("Cancel")),
        ),
        bt.WidgetButton(key="signatures", label=_("List all signatures"), icon=WBIcon.VIEW.icon),
        AbstractComplianceDocumentButtonConfig._get_compliance_document_button(),
    }


# COMPLIANCE FORM SIGNATURE
class ComplianceFormSignatureButtonConfig(ButtonViewConfig):
    CUSTOM_INSTANCE_BUTTONS = CUSTOM_LIST_INSTANCE_BUTTONS = {
        bt.ActionButton(
            method=RequestType.PATCH,
            identifiers=("wbcompliance:complianceformsignature",),
            key="signature",
            action_label=_("Signature"),
            title=_("Signature Compliance Form"),
            label=_("Signature"),
            icon=WBIcon.CONFIRM.icon,
            description_fields=_(
                """
            <p> <b>{{_compliance_form.form_type}} : <span>{{_compliance_form.title}}</span></b></p>
            <p>Version : {{version}}</p> <p><br/><br/>Do you want to sign <b>{{_compliance_form.title}}</b> ? </p>
            <p><span style='color:red'>This action is not reversible</span></p>
            """
            ),
            serializer=ComplianceFormSignatureModelSerializer,
            confirm_config=bt.ButtonConfig(label=_("Confirm")),
            cancel_config=bt.ButtonConfig(label=_("Cancel")),
            instance_display=create_simple_display([["remark"]]),
        ),
        AbstractComplianceDocumentButtonConfig._get_compliance_document_button(),
    }



import pytest
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import InternalUserFactory, UserFactory
from wbcore.contrib.authentication.models import User
from wbcore.test.utils import (
    get_data_from_factory,
    get_kwargs,
    get_model_factory,
    get_or_create_superuser,
)

from wbcompliance.factories import (
    ComplianceFormSectionFactory,
    ComplianceFormSignatureRuleFactory,
    ComplianceFormSignatureSectionFactory,
    ComplianceFormTypeFactory,
)
from wbcompliance.models import (
    ComplianceForm,
    ComplianceFormRule,
    ComplianceFormSection,
    ComplianceFormSignature,
    ComplianceFormType,
)
from wbcompliance.viewsets import (
    CFComplianceFormSignatureModelViewSet,
    ComplianceFormModelViewSet,
    ComplianceFormRuleViewSet,
    ComplianceFormSectionComplianceFormViewSet,
    ComplianceFormSectionRuleViewSet,
    ComplianceFormSignatureModelViewSet,
    ComplianceFormTypeViewSet,
    ComplianceTaskMatrixPandasViewSet,
)

from .mixins import UserTestMixin


@pytest.mark.django_db
class TestSpecificViewsets(UserTestMixin):
    @pytest.mark.parametrize(
        "mvs",
        [
            ComplianceFormModelViewSet,
            ComplianceFormSignatureModelViewSet,
            ComplianceFormTypeViewSet,
            ComplianceFormSectionComplianceFormViewSet,
            ComplianceFormSectionRuleViewSet,
        ],
    )
    def test_get_queryset(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=False)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize(
        "mvs, signature", [(ComplianceFormModelViewSet, False), (ComplianceFormSignatureModelViewSet, True)]
    )
    def test_option_request_declaration_on_honour(self, mvs, signature, compliance_form_factory):
        request = APIRequestFactory().options("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        complianceformtype = ComplianceFormTypeFactory()
        complianceformtype.type = ComplianceFormType.Type.FORM
        complianceformtype.save()
        if signature:
            cf = compliance_form_factory(form_type=complianceformtype)
            obj = factory(compliance_form=cf)
            sign_section = ComplianceFormSignatureSectionFactory(compliance_form_signature=obj)
            ComplianceFormSignatureRuleFactory(section=sign_section)
        else:
            obj = factory(form_type=complianceformtype)
            ComplianceFormSectionFactory(compliance_form=obj)
        obj.save()
        kwargs = get_kwargs(obj, mvs, request)
        kwargs["pk"] = obj.pk
        vs = mvs.as_view({"options": "options"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs, model, codename, is_queryset_not_empty",
        [
            (ComplianceFormTypeViewSet, ComplianceFormType, "view_complianceformtype", False),
            (ComplianceFormRuleViewSet, ComplianceFormRule, "view_complianceformrule", True),
            (ComplianceFormSectionComplianceFormViewSet, ComplianceFormSection, "view_complianceformsection", False),
            (ComplianceFormSectionRuleViewSet, ComplianceFormRule, "view_complianceformrule", False),
        ],
    )
    def test_option_request_no_admin_section_rules(self, mvs, model, codename, is_queryset_not_empty):
        request = APIRequestFactory().options("")
        request.user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        kwargs["pk"] = obj.pk
        content_type = ContentType.objects.get_for_model(model)
        permission = Permission.objects.get(
            codename=codename,
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        # permission check caching. -> refetch your object
        request.user.refresh_from_db()
        request.user = User.objects.get(pk=request.user.pk)

        view = mvs()
        view.setup(request, **kwargs)

        assert view.get_queryset().exists() == is_queryset_not_empty

    @pytest.mark.parametrize("mvs", [ComplianceFormModelViewSet])
    def test_option_request_no_admin_compliance_form(self, mvs, compliance_form_factory):
        request = APIRequestFactory().options("")
        request.user = InternalUserFactory(is_active=True)
        obj = compliance_form_factory.create()
        kwargs = get_kwargs(obj, mvs, request)

        permission_view = Permission.objects.get(
            codename="view_complianceform",
            content_type=ContentType.objects.get_for_model(ComplianceForm),
        )
        permission_select = Permission.objects.get(
            codename="select_complianceform",
            content_type=ContentType.objects.get_for_model(ComplianceForm),
        )
        permission_internal_user = Permission.objects.get(
            codename="is_internal_user",
        )
        request.user.user_permissions.add(permission_view)
        request.user.user_permissions.add(permission_select)
        request.user.user_permissions.add(permission_internal_user)

        request.user = User.objects.get(pk=request.user.pk)  # reset cached property and permission cache
        vs = mvs.as_view({"options": "options"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs, model, codename, is_allowed",
        [
            (ComplianceFormTypeViewSet, ComplianceFormType, "change_complianceformtype", False),
            (ComplianceFormRuleViewSet, ComplianceFormRule, "change_complianceformrule", False),
            (ComplianceFormSectionComplianceFormViewSet, ComplianceFormSection, "change_complianceformsection", False),
            (ComplianceFormSectionRuleViewSet, ComplianceFormRule, "change_complianceformrule", False),
        ],
    )
    def test_get_instance_endpoint_no_admin_section_rules(self, mvs, model, codename, is_allowed):
        user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        data = get_data_from_factory(obj, mvs, update=True, superuser=user)
        request = APIRequestFactory().put("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request)

        content_type = ContentType.objects.get_for_model(model)
        permission = Permission.objects.get(
            codename=codename,
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        # permission check caching. -> refetch your object
        request.user.refresh_from_db()
        request.user = User.objects.get(pk=request.user.pk)

        vs = mvs.as_view({"put": "update"})
        response = vs(request, **kwargs, pk=obj.pk)
        if is_allowed:
            assert response.status_code == status.HTTP_200_OK
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    @pytest.mark.parametrize("mvs", [ComplianceFormSignatureModelViewSet, CFComplianceFormSignatureModelViewSet])
    def test_get_create_endpoint_no_admin_compliance_form(self, mvs):
        user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        data = get_data_from_factory(obj, mvs, update=True, superuser=user)
        request = APIRequestFactory().post("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request)
        content_type = ContentType.objects.get_for_model(ComplianceFormSignature)
        permission = Permission.objects.get(
            codename="add_complianceformsignature",
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        request.user.refresh_from_db()
        request.user = User.objects.get(pk=request.user.pk)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    @pytest.mark.parametrize(
        "mvs, model, codename, is_allowed",
        [
            (ComplianceFormTypeViewSet, ComplianceFormType, "add_complianceformtype", False),
            (ComplianceFormRuleViewSet, ComplianceFormRule, "add_complianceformrule", False),
            (ComplianceFormSectionComplianceFormViewSet, ComplianceFormSection, "add_complianceformsection", False),
            (ComplianceFormSectionRuleViewSet, ComplianceFormRule, "add_complianceformrule", False),
        ],
    )
    def test_get_create_endpoint_no_admin_section_rules(self, mvs, model, codename, is_allowed):
        user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        data = get_data_from_factory(obj, mvs, update=True, superuser=user)
        request = APIRequestFactory().post("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request)

        content_type = ContentType.objects.get_for_model(model)
        permission = Permission.objects.get(
            codename=codename,
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        request.user.refresh_from_db()
        request.user = User.objects.get(pk=request.user.pk)

        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        if is_allowed:
            assert response.status_code == status.HTTP_201_CREATED
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    @pytest.mark.parametrize("mvs", [ComplianceFormSignatureModelViewSet, CFComplianceFormSignatureModelViewSet])
    def test_get_delete_endpoint_no_admin_compliance_form(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)

        content_type = ContentType.objects.get_for_model(ComplianceFormSignature)
        permission = Permission.objects.get(
            codename="delete_complianceformsignature",
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        request.user.refresh_from_db()

        # request.user = User.objects.get(pk=request.user.pk)
        vs = mvs.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    # TODO this should have never worked as Loic remove the delete endpoint excpliticlty, so the returned status code is 405 method not allowed and not 403.

    @pytest.mark.parametrize(
        "mvs, model, codename, is_allowed",
        [
            (ComplianceFormTypeViewSet, ComplianceFormType, "delete_complianceformtype", False),
            (ComplianceFormRuleViewSet, ComplianceFormRule, "delete_complianceformrule", False),
            (ComplianceFormSectionComplianceFormViewSet, ComplianceFormSection, "delete_complianceformsection", False),
            (ComplianceFormSectionRuleViewSet, ComplianceFormRule, "delete_complianceformrule", False),
        ],
    )
    def test_get_delete_endpoint_no_admin_section_rules(self, mvs, model, codename, is_allowed):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True)
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)

        content_type = ContentType.objects.get_for_model(model)
        permission = Permission.objects.get(
            codename=codename,
            content_type=content_type,
        )
        request.user.user_permissions.add(permission)
        request.user.refresh_from_db()

        # request.user = User.objects.get(pk=request.user.pk)
        vs = mvs.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        if is_allowed:
            assert response.status_code == status.HTTP_204_NO_CONTENT
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED

    @pytest.mark.parametrize(
        "mvs, no_employee", [(ComplianceFormModelViewSet, True), (ComplianceFormModelViewSet, False)]
    )
    def test_sendcomplianceformnotification(self, mvs, no_employee, user):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory()
        if not no_employee:
            obj.assigned_to.add(user.groups.first())
            obj.activationrequested()
            obj.active()

        response = mvs().sendcomplianceformnotification(request, pk=obj.pk)
        assert response
        assert response.data

    @pytest.mark.parametrize("mvs", [ComplianceFormSignatureModelViewSet, CFComplianceFormSignatureModelViewSet])
    def test_signature(self, mvs, user_factory):
        request = APIRequestFactory().get("")
        user = user_factory.create(is_superuser=False)
        superuser = user_factory.create(is_superuser=True)
        request.user = user

        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        obj = factory(person=user.profile)
        obj2 = factory(person=superuser.profile)

        response = mvs().signature(request, pk=obj.pk)
        response2 = mvs().signature(request, pk=obj2.pk)
        assert response
        assert response.data
        assert response2.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("mvs, changer", [(ComplianceFormModelViewSet, True), (ComplianceFormModelViewSet, False)])
    def test_active_compliance_form_viewset_admin(self, mvs, changer):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        factory = get_model_factory(mvs().get_serializer_class().Meta.model)
        if changer:
            obj = factory()
        else:
            obj = factory(changer=None)
        obj.activationrequested()
        obj.active()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        assert response.data.get("results")

    @pytest.mark.parametrize("mvs", [ComplianceTaskMatrixPandasViewSet])
    def test_pandasapiview(self, mvs, compliance_task_factory, compliance_task_instance_factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()

        compliance_task_instance_factory(task=compliance_task_factory())
        # kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.status_code == status.HTTP_200_OK, str(response.status_code) + " == 200"
        assert response.data, str(response.data) + " should not be empty"
        assert response.data.get("results"), str(response.data.get("results")) + " should not be empty"

    def test_queryset_filtering_for_user(self, user_factory, compliance_form_signature_factory):
        request = APIRequestFactory().get("")
        user = user_factory.create(is_superuser=False)
        form_signature = compliance_form_signature_factory.create()
        form_signature2 = compliance_form_signature_factory.create()

        request.user = user
        viewset = ComplianceFormSignatureModelViewSet(request=request)
        queryset = viewset.get_queryset()
        assert queryset.count() == 0

        form_signature.person = user.profile
        form_signature.save()
        queryset = viewset.get_queryset()
        assert set(queryset) == {form_signature}

        user.is_superuser = True
        user.save()
        request.user = user
        viewset = ComplianceFormSignatureModelViewSet(request=request)
        queryset = viewset.get_queryset()
        assert set(queryset) == {form_signature, form_signature2}
